from typing import Any

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.domain.policy import BaseUpdateUserPolicy
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp import Command


class UpdateUserPolicy(BaseUpdateUserPolicy):
    def __init__(self, abac_service: BaseAbacService, **kwargs: Any):
        super().__init__(abac_service, **kwargs)

        app_impl: AppImplDetails = abac_service.app.impl
        self.user_class: type[model.User] = app_impl.get_mapped_class(model.User)
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

    def is_allowed(self, cmd: Command) -> bool:
        user: model.User | None = cmd.user  # type: ignore[assignment]
        if user is None or user.id is None:
            return False
        roles = user.roles
        is_root = len(roles.intersection(self.role_set_map[enum.RoleSet.ROOT])) > 0

        tgt_user: model.User
        if isinstance(cmd, command.InviteUserCommand):
            if user.key == cmd.key:
                # User cannot invite themselves
                return False
            tgt_user = self.user_class(**cmd.model_dump())
            tgt_roles_union = set(tgt_user.roles)
            is_organization_update = False
        elif isinstance(cmd, command.UpdateUserCommand):
            tgt_user = self.abac_service.app.user_manager.retrieve_user_by_id(
                cmd.tgt_user_id
            )  # type: ignore[assignment]
            tgt_roles_union = set(tgt_user.roles)
            if cmd.roles is not None:
                tgt_roles_union.update(cmd.roles)
            is_organization_update = (
                cmd.organization_id is not None
                and cmd.organization_id != tgt_user.organization_id
            )
        else:
            raise NotImplementedError

        # Handle ROOT, >=APP_ADMIN, <ORG_ADMIN
        if is_root:
            # Root user can invite/update anyone with any permissions
            return True
        tgt_user.roles = tgt_roles_union
        if len(roles.intersection(self.role_set_map[enum.RoleSet.GE_APP_ADMIN])) > 0:
            # APP_ADMIN users and above can invite/update anyone with less permissions
            # (so not another APP_ADMIN), and only if the new set of permissions is also
            # less than their own permissions
            return self._has_more_permissions(user, tgt_user)
        if not roles.intersection(self.role_set_map[enum.RoleSet.GE_ORG_ADMIN]):
            # Only ORG_ADMIN users and above can invite/update users
            return False

        # ORG_ADMIN users can invite/update users of the organization(s) they are admin
        # of, but only if they have less permissions (so not another ORG_ADMIN), and
        # only if the new set of permissions is also less than their own permissions.
        # They cannot update the organization.
        if not tgt_roles_union.issubset(self.role_set_map[enum.RoleSet.LT_ORG_ADMIN]):
            # Target user has at least one role that is not under ORG_ADMIN
            return False
        if is_organization_update:
            # User cannot update the organization
            return False
        user_admin_organization_ids = (
            self.abac_service.retrieve_organizations_under_admin(
                command.RetrieveOrganizationsUnderAdminCommand(user=user)
            )
        )
        if tgt_user.organization_id not in user_admin_organization_ids:
            # User is not admin of the organization(s) the target user is part of
            return False
        return self._has_more_permissions(user, tgt_user)

    def _has_more_permissions(self, user: model.User, tgt_user: model.User) -> bool:
        permissions = self.abac_service.app.user_manager.retrieve_user_permissions(user)
        tgt_permissions = self.abac_service.app.user_manager.retrieve_user_permissions(
            tgt_user
        )
        return tgt_permissions.issubset(permissions) and len(permissions) > len(
            tgt_permissions
        )



# pylint: disable=wildcard-import, unused-import
# because this is a package, and imported as such in other modules
from __future__ import annotations

import datetime
import uuid
from enum import Enum

import ulid

from gen_epix.commondb.domain.enum import RoleSet as RoleSet


class TimestampFactory(Enum):
    DATETIME_NOW = lambda: datetime.datetime.now()


class IdFactory(Enum):
    UUID4 = uuid.uuid4
    ULID = lambda: ulid.api.new().uuid


class ServiceType(Enum):
    AUTH = "AUTH"
    ORGANIZATION = "ORGANIZATION"
    SYSTEM = "SYSTEM"
    RBAC = "RBAC"
    ABAC = "ABAC"
    SEQ = "SEQ"
    FILE = "FILE"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"


class Role(Enum):
    ROOT = "SEQDB_ROOT"
    APP_ADMIN = "SEQDB_APP_ADMIN"
    ORG_ADMIN = "SEQDB_ORG_ADMIN"
    REFDATA_ADMIN = "SEQDB_REFDATA_ADMIN"
    ORG_USER = "SEQDB_ORG_USER"
    GUEST = "SEQDB_GUEST"
    ROLE1 = "SEQDB_ROLE1"


class TreeAlgorithm(Enum):
    # See https://en.wikipedia.org/wiki/Hierarchical_clustering
    SLINK = "SLINK"  # Single linkage clustering
    CLINK = "CLINK"  # Complete linkage clustering
    UPGMA = "UPGMA"  # Unweighted average linkage clustering
    WPGMA = "WPGMA"  # Weighted average linkage clustering
    UPGMC = "UPGMC"  # Centroid linkage clustering
    WPGMC = "WPGMC"  # Median linkage clustering
    VERSATILE = "VERSATILE"  # Versatile linkage clustering
    MISSQ = "MISSQ"  # Ward linkage, Minimum Increase of Sum of Squares
    MNSSQ = "MNSSQ"  # Minimum Error Sum of Squares
    MIVAR = "MIVAR"  # Minimum Increase in Variance
    MNVAR = "MNVAR"  # Minimum Variance
    MINI_MAX = "MINI_MAX"  # Mini-Max linkage
    HAUSDORFF = "HAUSDORFF"  # Hausdorff linkage
    MIN_SUM_MEDOID = "MIN_SUM_MEDOID"  # Minimum Sum Medoid linkage
    MIN_SUM_INCREASE_MEDOID = (
        "MIN_SUM_INCREASE_MEDOID"  # Minimum Sum Increase Medoid linkage
    )
    MEDOID = "MEDOID"  # Medoid linkage
    MIN_ENERGY = "MIN_ENERGY"  # Minimum energy clustering
    FITCH_MARGOLIASH = "FITCH_MARGOLIASH"  # Fitch–Margoliash
    MAX_PARSIMONY = "MAX_PARSIMONY"  # Maximum parsimony
    ML = "ML"  # Maximum likelihood
    BAYESIAN_INFERENCE = "BAYESIAN_INFERENCE"  # Bayesian inference
    MIN_SPANNING = "MIN_SPANNING"  # Minimum spanning
    NJ = "NJ"  # Neighbor joining


class TreeAlgorithmSet(Enum):
    HIERARCHICAL_CLUSTERING = frozenset(
        {
            TreeAlgorithm.SLINK,
            TreeAlgorithm.CLINK,
            TreeAlgorithm.UPGMA,
            TreeAlgorithm.WPGMA,
            TreeAlgorithm.UPGMC,
            TreeAlgorithm.WPGMC,
            TreeAlgorithm.VERSATILE,
            TreeAlgorithm.MISSQ,
            TreeAlgorithm.MNSSQ,
            TreeAlgorithm.MIVAR,
            TreeAlgorithm.MNVAR,
            TreeAlgorithm.MINI_MAX,
            TreeAlgorithm.HAUSDORFF,
            TreeAlgorithm.MIN_SUM_MEDOID,
            TreeAlgorithm.MIN_SUM_INCREASE_MEDOID,
            TreeAlgorithm.MEDOID,
            TreeAlgorithm.MIN_ENERGY,
            TreeAlgorithm.FITCH_MARGOLIASH,
        }
    )
    NETWORK = frozenset({TreeAlgorithm.MIN_SPANNING})
    NJ = frozenset({TreeAlgorithm.NJ})
    PHYLOGENETIC_INFERENCE = frozenset(
        {
            TreeAlgorithm.MAX_PARSIMONY,
            TreeAlgorithm.ML,
            TreeAlgorithm.BAYESIAN_INFERENCE,
        }
    )
    DISTANCE_BASED = frozenset(
        {
            TreeAlgorithm.SLINK,
            TreeAlgorithm.CLINK,
            TreeAlgorithm.UPGMA,
            TreeAlgorithm.WPGMA,
            TreeAlgorithm.UPGMC,
            TreeAlgorithm.WPGMC,
            TreeAlgorithm.VERSATILE,
            TreeAlgorithm.MISSQ,
            TreeAlgorithm.MNSSQ,
            TreeAlgorithm.MIVAR,
            TreeAlgorithm.MNVAR,
            TreeAlgorithm.MINI_MAX,
            TreeAlgorithm.HAUSDORFF,
            TreeAlgorithm.MIN_SUM_MEDOID,
            TreeAlgorithm.MIN_SUM_INCREASE_MEDOID,
            TreeAlgorithm.MEDOID,
            TreeAlgorithm.MIN_ENERGY,
            TreeAlgorithm.FITCH_MARGOLIASH,
            TreeAlgorithm.MIN_SPANNING,
            TreeAlgorithm.NJ,
        }
    )


class Protocol(Enum):
    SEQUENCING = "SEQUENCING"
    LOCUS_DETECTION = "LOCUS_DETECTION"
    ALIGNMENT = "ALIGNMENT"
    TAXONOMY = "TAXONOMY"
    PCR = "PCR"
    AST = "AST"
    CLASSIFICATION = "CLASSIFICATION"
    SEQUENCE_DISTANCE = "SEQUENCE_DISTANCE"


class TaxonRank(Enum):
    NO_RANK = "NO_RANK"
    DOMAIN = "DOMAIN"
    SUPERKINGDOM = "SUPERKINGDOM"
    KINGDOM = "KINGDOM"
    SUBKINGDOM = "SUBKINGDOM"
    PHYLUM = "PHYLUM"
    SUBPHYLUM = "SUBPHYLUM"
    SUPERCLASS = "SUPERCLASS"
    CLASS = "CLASS"
    SUBCLASS = "SUBCLASS"
    INFRACLASS = "INFRACLASS"
    ORDER = "ORDER"
    SUBORDER = "SUBORDER"
    FAMILY = "FAMILY"
    SUBFAMILY = "SUBFAMILY"
    GENUS = "GENUS"
    SUBGENUS = "SUBGENUS"
    SPECIES_GROUP = "SPECIES_GROUP"
    SPECIES_SUBGROUP = "SPECIES_SUBGROUP"
    SPECIES = "SPECIES"
    SUBSPECIES = "SUBSPECIES"
    STRAIN = "STRAIN"
    CLADE = "CLADE"
    TRIBE = "TRIBE"


class QualityControlResult(Enum):
    PASS = "PASS"
    WARN = "WARN"
    FAIL = "FAIL"

    def is_usable(self) -> bool:
        return self in {QualityControlResult.PASS, QualityControlResult.WARN}


class SeqFormat(Enum):
    HASH_ONLY = "HASH_ONLY"  # Only the hash code of the sequence is known or stored
    STR_DNA5 = "STR_DNA5"  # String of ACTGN


class AlignmentFormat(Enum):
    CIGAR = "CIGAR"


class AlleleProfileFormat(Enum):
    SORTED_ALLELE_IDS = "SORTED_ALLELE_IDS"


class SnpProfileFormat(Enum):
    REF_ALN_SEQ = "REF_ALN_SEQ"


class KmerProfileFormat(Enum):
    KMER_PROFILE_FORMAT1 = "KMER_PROFILE_FORMAT1"


class SeqClassificationFormat(Enum):
    SEQ_CLASSIFICATION_FORMAT1 = "SEQ_CLASSIFICATION_FORMAT1"


class TaxonomyFormat(Enum):
    TAXONOMY_FORMAT1 = "TAXONOMY_FORMAT1"


class PcrResultFormat(Enum):
    PCR_RESULT_FORMAT1 = "PCR_RESULT_FORMAT1"


class AstResultFormat(Enum):
    AST_RESULT_FORMAT1 = "AST_RESULT_FORMAT1"


class SeqDistanceProtocolType(Enum):
    ALLELE_HAMMING = "ALLELE_HAMMING"
    SNP_HAMMING = "SNP_HAMMING"
    KMER_EUCLIDEAN = "KMER_EUCLIDEAN"
    OTHER = "OTHER"


class SeqDistanceProtocolTypeSet(Enum):
    ALLELE_BASED = frozenset({SeqDistanceProtocolType.ALLELE_HAMMING})
    SNP_BASED = frozenset({SeqDistanceProtocolType.SNP_HAMMING})
    KMER_BASED = frozenset({SeqDistanceProtocolType.KMER_EUCLIDEAN})


class SeqDistanceResultFormat(Enum):
    SEQ_DISTANCE_RESULT_FORMAT1 = "SEQ_DISTANCE_RESULT_FORMAT1"


class SeqDistanceFormat(Enum):
    SEQ_ID_DISTANCE_DICT = "SEQ_ID_DISTANCE_DICT"
    PROFILE_ID_DISTANCE_DICT = "PROFILE_ID_DISTANCE_DICT"


class SequenceFormat(Enum):
    FASTA = "FASTA"
    FASTQ = "FASTQ"



"""
Advanced streaming pipeline with backpressure handling and async support.
"""

import asyncio
from collections import deque
from collections.abc import Callable, Iterator
from typing import Any

from gen_epix.transform.pipeline import Pipeline
from gen_epix.transform.transform_result import TransformResult


class StreamingPipeline:
    """Advanced streaming pipeline with backpressure handling."""

    def __init__(
        self,
        pipeline: Pipeline,
        buffer_size: int = 1000,
        error_threshold: float = 0.1,
    ):
        self.pipeline = pipeline
        self.buffer_size = buffer_size
        self.error_threshold = error_threshold
        self.error_count = 0
        self.total_count = 0
        self.buffer: deque[Any] = deque(maxlen=buffer_size)

    def process_stream_async(
        self,
        stream: Iterator[Any],
        on_success: Callable[[TransformResult], None] | None = None,
        on_error: Callable[[TransformResult], None] | None = None,
    ) -> Iterator[TransformResult]:
        """Process stream asynchronously with callbacks."""

        for obj in stream:
            results = list(self.pipeline.process_stream(iter([obj])))

            for result in results:
                self.total_count += 1

                if result.success:
                    if on_success:
                        on_success(result)
                else:
                    self.error_count += 1
                    if on_error:
                        on_error(result)

                    # Check error threshold
                    error_rate = self.error_count / self.total_count
                    if error_rate > self.error_threshold:
                        raise RuntimeError(
                            f"Error rate {error_rate:.2%} exceeds threshold"
                        )

                yield result

    def collect_errors(
        self, stream: Iterator[Any]
    ) -> tuple[list[Any], list[TransformResult]]:
        """Collect successful and failed transformations separately."""
        successes: list[Any] = []
        errors: list[TransformResult] = []

        for result in self.process_stream_async(stream):
            if result.success:
                successes.append(result.transformed_object)
            else:
                errors.append(result)

        return successes, errors

    async def process_stream_async_coroutine(
        self, stream: Iterator[Any], batch_size: int = 100
    ) -> list[TransformResult]:
        """Process stream asynchronously using coroutines."""
        results: list[TransformResult] = []
        batch: list[Any] = []

        for obj in stream:
            batch.append(obj)

            if len(batch) >= batch_size:
                batch_results = await self._process_batch_async(batch)
                results.extend(batch_results)
                batch = []

        # Process remaining items
        if batch:
            batch_results = await self._process_batch_async(batch)
            results.extend(batch_results)

        return results

    async def _process_batch_async(self, batch: list[Any]) -> list[TransformResult]:
        """Process a batch of objects asynchronously."""
        tasks = [self._process_single_async(obj) for obj in batch]
        return await asyncio.gather(*tasks)

    async def _process_single_async(self, obj: Any) -> TransformResult:
        """Process a single object asynchronously."""
        # Run in thread pool to avoid blocking
        loop = asyncio.get_event_loop()
        results = await loop.run_in_executor(
            None, lambda: list(self.pipeline.process_stream(iter([obj])))
        )
        return (
            results[0]
            if results
            else TransformResult(
                success=False,
                original_object=obj,
                error=Exception("No results returned"),
                transformer_name="StreamingPipeline",
            )
        )



from __future__ import annotations

import abc
import uuid
from collections.abc import Hashable, Iterable
from typing import Any, ClassVar, Self

from pydantic import BaseModel as PydanticBaseModel
from pydantic import (
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    model_validator,
)

from gen_epix.fastapp import exc
from gen_epix.fastapp.domain.entity import Entity
from gen_epix.fastapp.enum import (
    CrudOperation,
    CrudOperationSet,
    PermissionType,
    PermissionTypeSet,
)
from gen_epix.filter.base import Filter


class Model(PydanticBaseModel):
    NAME: ClassVar[str | None] = None
    ENTITY: ClassVar[Entity | None] = None


class User(PydanticBaseModel):
    id: Hashable | None = Field(
        default_factory=uuid.uuid4,
        description="The ID of the user. This can be the key of the user (see get_key method), or a separate ID.",
    )

    def get_key(self) -> str:
        """
        Get the key of the user. The key is used to identify the user across systems,
        e.g. as a claim a in security token. This can be the email or any other unique
        identifier. Override this method to use retrieve the key in question, if
        different from the ID of the user.
        """
        return str(self.id)

    # @field_serializer("id", mode="plain")
    # def _serialize_id(self, value: Hashable) -> str:
    #     return str(value)


class Permission(PydanticBaseModel, frozen=True):
    """
    Implements a permission as a combination of (command_name, permission_type).
    The command_name is a string rather than the class of the command, to avoid
    issues with serialization such as for persistence and for API requests/responses.
    """

    _NAME_DELIMITER: ClassVar[str] = "_"

    command_name: str
    permission_type: PermissionType

    @computed_field
    def name(self) -> str:
        return f"{self.command_name}{Permission._NAME_DELIMITER}{self.permission_type.value}"

    @computed_field
    def sort_key(self) -> tuple[str, int]:
        permission_type_map = {
            PermissionType.EXECUTE: 0,
            PermissionType.CREATE: 1,
            PermissionType.READ: 2,
            PermissionType.UPDATE: 3,
            PermissionType.DELETE: 4,
        }
        return self.command_name, permission_type_map[self.permission_type]

    def __eq__(self, permission: object) -> bool:
        # TODO: Investigate why two objs of this class with the same values are
        # not equal without overriding __eq__
        if not isinstance(permission, Permission):
            return False
        return (
            self.name == permission.name
            and self.permission_type == permission.permission_type
        )

    def __repr__(self) -> str:
        return f"({self.command_name},{self.permission_type.value})"

    @field_serializer("permission_type", mode="plain")
    def _serialize_permission_type(self, value: PermissionType) -> str:
        return value.value


class Policy(abc.ABC):
    def get_is_denied_exception(self) -> type[Exception]:
        return exc.UnauthorizedAuthError

    # Not an abstract method since it is not always needed
    def is_allowed(self, cmd: Command) -> bool:
        raise NotImplementedError

    # Not an abstract method since it is not always needed
    def get_content(self, cmd: Command) -> Any:
        raise NotImplementedError

    # Not an abstract method since it is not always needed
    def get_content_return_type(self, cmd: Command) -> type:
        raise NotImplementedError

    # Not an abstract method since it is not always needed
    def filter(self, cmd: Command, retval: Any) -> Any:
        raise NotImplementedError


class Command(PydanticBaseModel):
    PERMISSION_TYPE_SET: ClassVar[PermissionTypeSet] = PermissionTypeSet.E
    NAME: ClassVar[str | None] = None

    _PERMISSIONS: ClassVar[frozenset[Permission] | None] = None

    id: Hashable = Field(
        default_factory=uuid.uuid4, description="The ID of the command obj"
    )
    user: User | None = None
    _policies: list[Policy] = PrivateAttr(default_factory=list)

    # @field_serializer("id", mode="plain")
    # def _serialize_id(self, value: Hashable) -> str | None:
    #     return serialize_id(value)


class CrudCommand(Command):
    PERMISSION_TYPE_SET: ClassVar[PermissionTypeSet] = PermissionTypeSet.CRUD
    MODEL_CLASS: ClassVar[type[Model]] = Model

    operation: CrudOperation = Field(description="The CRUD operation to perform.")
    obj_ids: Hashable | list[Hashable] | None = Field(
        default=None,
        description="The identifier(s) of the object(s) to operate on. Must be set to a single identifier for read, delete or exists one operations, and to a list of identifiers for read, delete or exists some operations. Otherwise must be None.",
    )
    objs: Model | list[Model] | None = Field(
        default=None,
        description="The object(s) to operate on. Must be set to a single object for create or update one operations, and to a list of objects for create or update some operations. Otherwise must be None.",
    )
    query_filter: Filter | None = Field(
        default=None,
        description="Optional filter to apply to the results of a read or delete all operation, thereby effectively applying a query instead of reading or deleting all. Must be None for all other operations.",
    )
    access_filter: Filter | None = Field(
        default=None,
        description="Optional filter to apply object-level access control. For a read or delete all operation, it filters the results just as the query_filter does and when both are provided only object that match both filters will be returned or deleted. For any other operation, an unauthorized exception is raised if the provided objects do not match the filter.",
    )
    props: dict[str, Any] = Field(
        default={},
        description="Additional properties to pass to the command and which can be used by custom implementations.",
    )

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        operation = self.operation
        obj_ids = self.obj_ids
        objs = self.objs
        if obj_ids is None:
            if (
                operation not in CrudOperationSet.ANY_ALL.value
                and operation not in CrudOperationSet.WRITE.value
            ):
                raise ValueError(
                    f"Invalid operation for obj_ids=None: {operation.value}"
                )
        elif isinstance(obj_ids, Iterable) and not isinstance(obj_ids, Model):
            if operation not in CrudOperationSet.NON_CREATE_SOME.value:
                raise ValueError(
                    f"Invalid operation for obj_ids=list: {operation.value}"
                )
        else:
            if operation not in CrudOperationSet.NON_CREATE_ONE.value:
                raise ValueError(
                    f"Invalid operation for obj_ids=obj_id: {operation.value}"
                )
        if objs is None:
            if operation in CrudOperationSet.WRITE.value:
                raise ValueError(
                    f"Invalid operation for objects=None: {operation.value}"
                )
        elif isinstance(objs, Iterable) and not isinstance(objs, Model):
            if operation not in CrudOperationSet.WRITE_SOME.value:
                raise ValueError(f"Invalid operation for objs=list: {operation.value}")
        else:
            if operation not in CrudOperationSet.WRITE_ONE.value or not isinstance(
                objs, Model
            ):
                raise ValueError(
                    f"Invalid operation for objs=object: {operation.value}"
                )
        if self.query_filter is not None:
            if operation not in CrudOperationSet.ANY_ALL.value:
                raise ValueError(
                    f"Invalid operation for query_filter not None: {operation.value}"
                )
        return self

    def get_obj_ids(
        self, as_set: bool = False
    ) -> list[Hashable | None] | set[Hashable] | None:
        """
        Get the object IDs, either from the obj_ids field or from the objs field. In
        the latter case, the IDs are extracted from the objects using the entity's ID
        field name. In case the command has obj_ids=None, None is returned. If
        as_set=True, a set of IDs and excluding None is returned where otherwise a list
        would be returned.
        """
        if self.obj_ids is not None:
            # Command has obj_ids and cannot have objs
            if as_set:
                return (
                    set(self.obj_ids)
                    if isinstance(self.obj_ids, list)
                    else {self.obj_ids}
                )
            return self.obj_ids if isinstance(self.obj_ids, list) else [self.obj_ids]
        elif self.objs is not None:
            # Command has objs and cannot have obj_ids
            entity = self.MODEL_CLASS.ENTITY
            if entity is None:
                raise exc.InitializationServiceError(
                    f"Entity not set for model {self.MODEL_CLASS.__name__}"
                )
            id_field_name = entity.get_id_field_name()
            if isinstance(self.objs, list):
                if as_set:
                    retval = {getattr(obj, id_field_name) for obj in self.objs}
                    retval.discard(None)
                    return retval
                return [getattr(obj, id_field_name) for obj in self.objs]
            else:
                if as_set:
                    retval = {getattr(self.objs, id_field_name)}
                    retval.discard(None)
                    return retval
                return [getattr(self.objs, id_field_name)]
        # Command has neither obj_ids nor objs
        return None

    def get_objs(self) -> list[Model] | None:
        """
        Get the objects as a list, or None if no objects.
        """
        if self.objs is not None:
            return self.objs if isinstance(self.objs, list) else [self.objs]
        return None


class UpdateAssociationCommand(Command):
    ASSOCIATION_CLASS: ClassVar[type[Model]] = Model
    LINK_FIELD_NAME1: ClassVar[str] = ""
    LINK_FIELD_NAME2: ClassVar[str] = ""

    obj_id1: Hashable | None = Field(
        default=None,
        description="The ID of the instance of the first entity in the many-to-many association.",
    )
    obj_id2: Hashable | None = Field(
        default=None,
        description="The ID of the instance of the second entity in the many-to-many association.",
    )
    association_objs: list[Model] = Field(
        default=[],
        description="The association objects, linking the first (field LINK_FIELD_NAME1) to the second (field LINK_FIELD_NAME2) instance.",
    )
    props: dict[str, Any] = {}

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        obj_id1 = self.obj_id1
        obj_id2 = self.obj_id2
        association_objs = self.association_objs
        if obj_id1 and obj_id2:
            raise exc.DomainException(
                f"Invalid state: obj_id1 and obj_id2 are both present"
            )
        if association_objs:
            if obj_id1 and not all(
                getattr(obj, self.LINK_FIELD_NAME1) == obj_id1
                for obj in association_objs
            ):
                raise exc.DomainException(
                    f"Invalid state: obj_id1 and association_objs not matching"
                )
            if obj_id2 and not all(
                getattr(obj, self.LINK_FIELD_NAME2) == obj_id2
                for obj in association_objs
            ):
                raise exc.DomainException(
                    f"Invalid state: obj_id2 and association_objs not matching"
                )
        else:
            if not obj_id1 and not obj_id2:
                raise exc.DomainException(
                    f"Invalid state: association_objs, obj_id1 and obj_id2 all empty"
                )
        return self


class Role(PydanticBaseModel):
    name: str
    permissions: set[Permission]



"""
Field transformer implementation.
"""

from collections.abc import Callable, Hashable
from typing import Any

from gen_epix.transform.adapter import ObjectAdapter
from gen_epix.transform.transformer import Transformer


class FieldTransformer(Transformer):
    """Transform a specific field in an object."""

    def __init__(
        self,
        field_name: Hashable,
        transform_fn: Callable[[Any], Any],
        name: str | None = None,
    ):
        super().__init__(name)
        self.field_name = field_name
        self.transform_fn = transform_fn

    def transform(self, obj: ObjectAdapter) -> ObjectAdapter:
        """Transform the specified field if it exists."""
        if obj.has_key(self.field_name):
            current_value = obj.get(self.field_name)
            transformed_value = self.transform_fn(current_value)
            obj.set(self.field_name, transformed_value)
        return obj



# pylint: disable=useless-import-alias
from gen_epix.fastapp.api.crud_endpoint_generator import (
    CrudEndpointGenerator as CrudEndpointGenerator,
)
from gen_epix.fastapp.api.crud_endpoint_set import CrudEndpointSet as CrudEndpointSet



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar
from uuid import UUID

from pydantic import Field

from gen_epix.casedb.domain import enum
from gen_epix.commondb.domain.model.base import Model
from gen_epix.fastapp.domain import Entity, create_keys, create_links


class RegionSet(Model):
    """
    Set of regions that do not overlap geographically
    or otherwise did not exist at the same moment in time.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="region_sets",
        table_name="region_set",
        persistable=True,
        keys=create_keys({1: "code", 2: "name"}),
    )
    code: str = Field(description="The code of the region set.", max_length=255)
    name: str = Field(
        description="The name of the region set.",
        max_length=255,
    )
    region_code_as_label: bool = Field(
        description=(
            "Whether the region's code should be used as the label."
            " E.g. in case of postal code the code "
            "could be used instead of the name of the region."
        ),
    )


class RegionSetShape(Model):
    """
    Geographical shape representation for a region set.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="region_set_shapes",
        table_name="region_set_shape",
        persistable=True,
        keys=create_keys({1: ("region_set_id", "scale")}),
        links=create_links(
            {
                1: ("region_set_id", RegionSet, "region_set"),
            }
        ),
    )
    region_set_id: UUID = Field(description="The ID of the region set. FOREIGN KEY")
    region_set: RegionSet | None = Field(
        default=None, description="The region set to which the region belongs."
    )
    scale: float = Field(description="The scale of the shape representation.")
    geo_json: str = Field(
        description="The GeoJSON representation of the region set shape."
    )


class Region(Model):
    """
    Geographical representation of a region.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="regions",
        table_name="region",
        persistable=True,
        keys=create_keys(
            {
                1: ("region_set_id", "code"),
                # 2: ("region_set_id", "name"),
                # # postal codes in NL can have the same name
            }
        ),
        links=create_links(
            {
                1: ("region_set_id", RegionSet, "region_set"),
            }
        ),
    )
    region_set_id: UUID = Field(description="The ID of the region set. FOREIGN KEY")
    region_set: RegionSet | None = Field(
        default=None, description="The region set to which the region belongs."
    )
    code: str = Field(description="The code of the region.", max_length=255)
    name: str = Field(
        description="The name of the region.",
        max_length=255,
    )
    centroid_lat: float = Field(description="The latitude of the region's centroid.")
    centroid_lon: float = Field(description="The longitude of the region's centroid.")
    center_lat: float = Field(description="The latitude of the region's center.")
    center_lon: float = Field(description="The longitude of the region's center.")


class RegionRelation(Model):
    """
    Geographical relation between two regions.
    """

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="region_relations",
        table_name="region_relation",
        persistable=True,
        keys=create_keys({1: ("from_region_id", "to_region_id")}),
        links=create_links(
            {
                1: ("from_region_id", Region, "from_region"),
                2: ("to_region_id", Region, "to_region"),
            }
        ),
    )
    from_region_id: UUID = Field(description="The ID of the source region. FOREIGN KEY")
    from_region: Region | None = Field(default=None, description="The source region.")
    to_region_id: UUID = Field(description="The ID of the target region. FOREIGN KEY")
    to_region: Region | None = Field(default=None, description="The target region.")
    relation: enum.RegionRelationType = Field(
        description="The type of relation between the regions."
    )



# pylint: disable=useless-import-alias
from gen_epix.commondb.repositories.sa_model import Contact as Contact
from gen_epix.commondb.repositories.sa_model import DataCollection as DataCollection
from gen_epix.commondb.repositories.sa_model import (
    DataCollectionSet as DataCollectionSet,
)
from gen_epix.commondb.repositories.sa_model import (
    DataCollectionSetMember as DataCollectionSetMember,
)
from gen_epix.commondb.repositories.sa_model import IdentifierIssuer as IdentifierIssuer
from gen_epix.commondb.repositories.sa_model import Organization as Organization
from gen_epix.commondb.repositories.sa_model import (
    OrganizationAdminPolicy as OrganizationAdminPolicy,
)
from gen_epix.commondb.repositories.sa_model import OrganizationSet as OrganizationSet
from gen_epix.commondb.repositories.sa_model import (
    OrganizationSetMember as OrganizationSetMember,
)
from gen_epix.commondb.repositories.sa_model import Outage as Outage
from gen_epix.commondb.repositories.sa_model import RowMetadataMixin
from gen_epix.commondb.repositories.sa_model import Site as Site
from gen_epix.commondb.repositories.sa_model import User as User
from gen_epix.commondb.repositories.sa_model import UserInvitation as UserInvitation
from gen_epix.commondb.repositories.sa_model import (
    create_field_metadata,
    set_entity_repository_model_classes,
)
from gen_epix.seqdb.domain import DOMAIN, enum, model
from gen_epix.seqdb.repositories.sa_model.file import File as File
from gen_epix.seqdb.repositories.sa_model.seq import (
    AlignmentProtocol as AlignmentProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import Allele as Allele
from gen_epix.seqdb.repositories.sa_model.seq import AlleleAlignment as AlleleAlignment
from gen_epix.seqdb.repositories.sa_model.seq import AlleleProfile as AlleleProfile
from gen_epix.seqdb.repositories.sa_model.seq import (
    AssemblyProtocol as AssemblyProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import AstMeasurement as AstMeasurement
from gen_epix.seqdb.repositories.sa_model.seq import AstPrediction as AstPrediction
from gen_epix.seqdb.repositories.sa_model.seq import AstProtocol as AstProtocol
from gen_epix.seqdb.repositories.sa_model.seq import (
    KmerDetectionProtocol as KmerDetectionProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import KmerProfile as KmerProfile
from gen_epix.seqdb.repositories.sa_model.seq import (
    LibraryPrepProtocol as LibraryPrepProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import Locus as Locus
from gen_epix.seqdb.repositories.sa_model.seq import (
    LocusDetectionProtocol as LocusDetectionProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import LocusSet as LocusSet
from gen_epix.seqdb.repositories.sa_model.seq import LocusSetMember as LocusSetMember
from gen_epix.seqdb.repositories.sa_model.seq import PcrMeasurement as PcrMeasurement
from gen_epix.seqdb.repositories.sa_model.seq import PcrProtocol as PcrProtocol
from gen_epix.seqdb.repositories.sa_model.seq import RawSeq as RawSeq
from gen_epix.seqdb.repositories.sa_model.seq import ReadSet as ReadSet
from gen_epix.seqdb.repositories.sa_model.seq import RefAllele as RefAllele
from gen_epix.seqdb.repositories.sa_model.seq import RefSeq as RefSeq
from gen_epix.seqdb.repositories.sa_model.seq import RefSnp as RefSnp
from gen_epix.seqdb.repositories.sa_model.seq import RefSnpSet as RefSnpSet
from gen_epix.seqdb.repositories.sa_model.seq import RefSnpSetMember as RefSnpSetMember
from gen_epix.seqdb.repositories.sa_model.seq import Sample as Sample
from gen_epix.seqdb.repositories.sa_model.seq import Seq as Seq
from gen_epix.seqdb.repositories.sa_model.seq import SeqAlignment as SeqAlignment
from gen_epix.seqdb.repositories.sa_model.seq import SeqCategory as SeqCategory
from gen_epix.seqdb.repositories.sa_model.seq import SeqCategorySet as SeqCategorySet
from gen_epix.seqdb.repositories.sa_model.seq import (
    SeqClassification as SeqClassification,
)
from gen_epix.seqdb.repositories.sa_model.seq import (
    SeqClassificationProtocol as SeqClassificationProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import SeqDistance as SeqDistance
from gen_epix.seqdb.repositories.sa_model.seq import (
    SeqDistanceProtocol as SeqDistanceProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import SeqTaxonomy as SeqTaxonomy
from gen_epix.seqdb.repositories.sa_model.seq import (
    SnpDetectionProtocol as SnpDetectionProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import SnpProfile as SnpProfile
from gen_epix.seqdb.repositories.sa_model.seq import SubtypingScheme as SubtypingScheme
from gen_epix.seqdb.repositories.sa_model.seq import Taxon as Taxon
from gen_epix.seqdb.repositories.sa_model.seq import TaxonLocusLink as TaxonLocusLink
from gen_epix.seqdb.repositories.sa_model.seq import (
    TaxonomyProtocol as TaxonomyProtocol,
)
from gen_epix.seqdb.repositories.sa_model.seq import TaxonSet as TaxonSet
from gen_epix.seqdb.repositories.sa_model.seq import TaxonSetMember as TaxonSetMember
from gen_epix.seqdb.repositories.sa_model.seq import TreeAlgorithm as TreeAlgorithm
from gen_epix.seqdb.repositories.sa_model.seq import (
    TreeAlgorithmClass as TreeAlgorithmClass,
)

SA_MODELS_BY_SERVICE_TYPE: dict[enum.ServiceType, dict[type[model.Model], type]] = {
    enum.ServiceType.ABAC: {
        model.OrganizationAdminPolicy: OrganizationAdminPolicy,
    },
    enum.ServiceType.ORGANIZATION: {
        model.Contact: Contact,
        model.DataCollection: DataCollection,
        model.DataCollectionSet: DataCollectionSet,
        model.DataCollectionSetMember: DataCollectionSetMember,
        model.IdentifierIssuer: IdentifierIssuer,
        model.Organization: Organization,
        model.OrganizationSet: OrganizationSet,
        model.OrganizationSetMember: OrganizationSetMember,
        model.Site: Site,
        model.User: User,
        model.UserInvitation: UserInvitation,
    },
    enum.ServiceType.SYSTEM: {
        model.Outage: Outage,
    },
    enum.ServiceType.SEQ: {
        model.AlignmentProtocol: AlignmentProtocol,
        model.Allele: Allele,
        model.AlleleAlignment: AlleleAlignment,
        model.AlleleProfile: AlleleProfile,
        model.AssemblyProtocol: AssemblyProtocol,
        model.AstMeasurement: AstMeasurement,
        model.AstPrediction: AstPrediction,
        model.AstProtocol: AstProtocol,
        model.KmerDetectionProtocol: KmerDetectionProtocol,
        model.KmerProfile: KmerProfile,
        model.LibraryPrepProtocol: LibraryPrepProtocol,
        model.Locus: Locus,
        model.LocusDetectionProtocol: LocusDetectionProtocol,
        model.LocusSet: LocusSet,
        model.LocusSetMember: LocusSetMember,
        model.PcrMeasurement: PcrMeasurement,
        model.PcrProtocol: PcrProtocol,
        model.RawSeq: RawSeq,
        model.ReadSet: ReadSet,
        model.RefAllele: RefAllele,
        model.RefSeq: RefSeq,
        model.RefSnp: RefSnp,
        model.RefSnpSet: RefSnpSet,
        model.RefSnpSetMember: RefSnpSetMember,
        model.Sample: Sample,
        model.Seq: Seq,
        model.SeqAlignment: SeqAlignment,
        model.SeqCategory: SeqCategory,
        model.SeqCategorySet: SeqCategorySet,
        model.SeqClassification: SeqClassification,
        model.SeqClassificationProtocol: SeqClassificationProtocol,
        model.SeqDistance: SeqDistance,
        model.SeqDistanceProtocol: SeqDistanceProtocol,
        model.SeqTaxonomy: SeqTaxonomy,
        model.SnpDetectionProtocol: SnpDetectionProtocol,
        model.SnpProfile: SnpProfile,
        model.SubtypingScheme: SubtypingScheme,
        model.Taxon: Taxon,
        model.TaxonLocusLink: TaxonLocusLink,
        model.TaxonomyProtocol: TaxonomyProtocol,
        model.TaxonSet: TaxonSet,
        model.TaxonSetMember: TaxonSetMember,
        model.TreeAlgorithm: TreeAlgorithm,
        model.TreeAlgorithmClass: TreeAlgorithmClass,
    },
    enum.ServiceType.FILE: {
        model.File: File,
    },
}

FIELD_NAME_MAP: dict[type, dict[str, str]] = {}

set_entity_repository_model_classes(
    DOMAIN,
    SA_MODELS_BY_SERVICE_TYPE,
    RowMetadataMixin,
    field_name_map=FIELD_NAME_MAP,
)

SERVICE_METADATA_FIELDS, DB_METADATA_FIELDS, GENERATE_SERVICE_METADATA = (
    create_field_metadata(DOMAIN)
)



from typing import Any

from gen_epix.casedb.domain import command
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.commondb.policies import (
    ReadSelfResultsOnlyPolicy as CommonReadSelfResultsOnlyPolicy,
)


class ReadSelfResultsOnlyPolicy(CommonReadSelfResultsOnlyPolicy):
    def __init__(
        self,
        abac_service: BaseAbacService,
        **kwargs: Any,
    ):
        super().__init__(
            abac_service,
            **kwargs,
        )
        self.id_attr_by_command_class.update(
            {
                command.UserAccessCasePolicyCrudCommand: "user_id",
                command.UserShareCasePolicyCrudCommand: "user_id",
            }
        )



from collections.abc import Iterable
from uuid import UUID

import numpy as np

from gen_epix.fastapp.repositories import DictRepository
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.seqdb.domain import model
from gen_epix.seqdb.domain.repository import BaseSeqRepository


class SeqDictRepository(DictRepository, BaseSeqRepository):
    def get_distance_matrix_by_seq_ids(
        self,
        uow: BaseUnitOfWork,
        seq_distance_protocol_id: UUID,
        seq_ids: list[UUID],
    ) -> np.ndarray:
        raise NotImplementedError("Code to be converted to seqdb architecture")
        self.raise_on_duplicate_ids(seq_ids)
        seqs = self.read_some(model.SeqDistance, seq_ids)
        id_to_idx_map = {x.id: i for i, x in enumerate(seqs)}
        n = len(seqs)
        distance_matrix = np.empty((n, n))
        distance_matrix[:] = np.nan
        for i in range(n):
            for id_, distance in seqs[i].distances.items():
                if id_ not in id_to_idx_map:
                    continue
                distance_matrix[id_to_idx_map[id_], i] = distance
            distance_matrix[i, i] = 0
        return distance_matrix

    def retrieve_seq_fasta(
        self,
        uow: BaseUnitOfWork,
        seq_ids: list[UUID],
    ) -> Iterable[tuple[str, str]]:
        self.raise_on_duplicate_ids(seq_ids)

        for seq_id in seq_ids:
            seq: model.Seq = self.read_one(model.Seq, seq_id)  # type: ignore[assignment]
            raw_seq: model.RawSeq = self.read_one(  # type: ignore[assignment]
                model.RawSeq,
                seq.raw_seq_id,
            )
            yield (seq.id, raw_seq.seq)  # type: ignore[misc]



import datetime
import logging
import re
from collections.abc import Hashable
from enum import Enum
from pathlib import Path
from time import sleep
from typing import Any, TypeVar, cast
from uuid import UUID

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.base_env import BaseAppComposer
from gen_epix.commondb.config import BaseAppCfg
from gen_epix.commondb.domain import command, enum, model
from gen_epix.commondb.test.endpoint_test_client import EndpointTestClient
from gen_epix.commondb.test.util import set_log_level
from gen_epix.commondb.util import map_paired_elements
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.fastapp.model import Command

BASE_MODEL_TYPE = TypeVar("BASE_MODEL_TYPE", bound=model.Model)


class TestClient:

    DEFAULT_ROUTE_PREFIX = "/v1"

    MODEL_KEY_MAP: dict[type[model.Model], str | tuple[str, ...]] = {
        model.User: "name",
        model.UserInvitation: "email",
        model.Organization: "name",
        model.DataCollection: "name",
        model.OrganizationAdminPolicy: ("organization_id", "user_id"),
        model.Site: ("name"),
        model.Contact: ("name"),
    }

    def __init__(
        self,
        test_name: str,
        test_dir: Path,
        app_cfg: BaseAppCfg,
        app_composer: BaseAppComposer,
        verbose: bool = False,
        log_level: int = logging.ERROR,
        use_endpoints: bool = False,
        default_route_prefix: str | None = None,
        **kwargs: Any,
    ):
        # Set provided parameters
        self.test_name = test_name
        self.test_dir = test_dir
        self.app_cfg = app_cfg
        self.app_composer = app_composer
        self.default_route_prefix = default_route_prefix or self.DEFAULT_ROUTE_PREFIX
        self.log_level = log_level
        self.verbose = verbose

        # Get implementation details
        app_impl: AppImplDetails = app_composer.app.impl
        self.user_class: type[model.User] = app_impl.get_mapped_class(model.User)
        self.user_invitation_class: type[model.UserInvitation] = (
            app_impl.get_mapped_class(model.UserInvitation)
        )
        self.user_invitation_constraints_class: type[
            model.UserInvitationConstraints
        ] = app_impl.get_mapped_class(model.UserInvitationConstraints)
        self.organization_admin_policy_class: type[model.OrganizationAdminPolicy] = (
            app_impl.get_mapped_class(model.OrganizationAdminPolicy)
        )
        self.user_crud_command_class: type[command.UserCrudCommand] = (
            app_impl.get_mapped_class(command.UserCrudCommand)
        )
        self.user_invitation_crud_command_class: type[
            command.UserInvitationCrudCommand
        ] = app_impl.get_mapped_class(command.UserInvitationCrudCommand)
        self.organization_admin_policy_crud_command_class: type[
            command.OrganizationAdminPolicyCrudCommand
        ] = app_impl.get_mapped_class(command.OrganizationAdminPolicyCrudCommand)
        self.retrieve_invite_user_constraints_command_class: type[
            command.RetrieveInviteUserConstraintsCommand
        ] = app_impl.get_mapped_class(command.RetrieveInviteUserConstraintsCommand)
        self.invite_user_command_class: type[command.InviteUserCommand] = (
            app_impl.get_mapped_class(command.InviteUserCommand)
        )
        self.register_invited_user_command_class: type[
            command.RegisterInvitedUserCommand
        ] = app_impl.get_mapped_class(command.RegisterInvitedUserCommand)
        self.retrieve_organization_admin_name_emails_command_class: type[
            command.RetrieveOrganizationAdminNameEmailsCommand
        ] = app_impl.get_mapped_class(
            command.RetrieveOrganizationAdminNameEmailsCommand
        )
        self.update_user_command_class: type[command.UpdateUserCommand] = (
            app_impl.get_mapped_class(command.UpdateUserCommand)
        )
        self.role_map = app_impl.role_map
        self.rev_role_map = app_impl.rev_role_map
        self.role_set_map = app_impl.role_set_map
        self.role_permissions_map = app_impl.role_permissions_map
        self.app = self.app_composer.app
        self.cfg = self.app.cfg
        self.services = app_impl.services
        self.repositories = app_impl.repositories

        # Set log level
        TestClient._set_log_level(app_cfg, log_level)

        # Set additional parameters
        self.root_role: str = self.role_map[enum.Role.ROOT]
        self.guest_role: str = self.role_map[enum.Role.GUEST]
        self.db: dict[type[model.Model], dict[Hashable, model.Model]] = {}
        self.props: dict = {}
        self.use_endpoints = use_endpoints
        self.default_route_prefix = default_route_prefix or self.DEFAULT_ROUTE_PREFIX
        self.endpoint_test_client: EndpointTestClient | None = kwargs.pop(
            "endpoint_test_client"
        )
        self.app_last_handled_exception: dict | None = kwargs.pop(
            "app_last_handled_exception"
        )
        if self.use_endpoints:
            if not self.endpoint_test_client:
                raise ValueError(
                    "Endpoint test client not provided while use_endpoints=True"
                )
            if not self.app_last_handled_exception:
                raise ValueError(
                    "App last handled exception not provided while use_endpoints=True"
                )

        # Store remainder of kwargs
        self.props = kwargs

    def handle(
        self,
        cmd: Command,
        return_response: bool = False,
        use_endpoint: bool | None = None,
        route_prefix: str | None = None,
        **kwargs: Any,
    ) -> Any:
        use_endpoint = use_endpoint if use_endpoint is not None else self.use_endpoints
        if use_endpoint:
            assert self.app_last_handled_exception is not None
            assert self.endpoint_test_client is not None

            route_prefix = route_prefix or self.default_route_prefix
            previous_exception_id = self.app_last_handled_exception["id"]
            retval, response = self.endpoint_test_client.handle(
                cmd,
                return_response=True,
                route_prefix=route_prefix,
                **kwargs,
            )

            # Check if an exception was raised before generating the HTTP response, and
            # if so, raise it
            exception_id = self.app_last_handled_exception["id"]
            if exception_id != previous_exception_id:
                exception = self.app_last_handled_exception["exception"]
                raise exception

            if return_response:
                return retval, response
            return retval

        else:
            return self.app.handle(cmd)

    def retrieve_user_by_key(self, user_key: str) -> model.User:
        user: model.User = self.app.user_manager.retrieve_user_by_key(user_key)  # type: ignore[assignment]
        return user

    def create_organization(
        self, user_or_str: str | model.User, organization_name: str
    ) -> model.Organization:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        organization = self.app.handle(
            command.OrganizationCrudCommand(
                user=user,
                operation=CrudOperation.CREATE_ONE,
                objs=model.Organization(
                    name=organization_name, legal_entity_code=organization_name
                ),
            )
        )
        retval: model.Organization = self._set_obj(organization)  # type: ignore[assignment]
        return retval

    def invite_and_register_user(
        self,
        user_or_str: str | model.User,
        user_name: str,
        set_dummy_organization: bool = False,
        set_dummy_token: bool = False,
    ) -> model.User:
        root_user: model.User = self.get_root_user()
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        m = re.match(r"^(.*?)(\d+)_(\d+)$", user_name.lower())
        if not m:
            raise ValueError(f"Invalid user name {user_name}")
        role = [
            y for x, y in self.role_map.items() if x.name.lower() == m.group(1).lower()
        ][0]
        organization_name = "org" + m.group(2)
        organization_id: UUID
        if organization_name not in self.db[model.Organization]:
            if set_dummy_organization:
                organization_id = self.generate_id()
            else:
                raise ValueError(f"Organization {organization_name} not found")
        else:
            organization_id = self.db[model.Organization][organization_name].id  # type: ignore[assignment]
        user_invitation: model.UserInvitation = self.handle(
            self.invite_user_command_class(
                user=user,
                key=f"{user_name}@{organization_name}.org",
                email=f"{user_name}@{organization_name}.org",
                roles={role},
                organization_id=organization_id,
            )
        )
        if set_dummy_token:
            user_invitation.token = str(self.generate_id())
        tgt_user: model.User = self.handle(
            self.register_invited_user_command_class(
                user=self.user_class(
                    key=f"{user_name}@{organization_name}.org",
                    email=f"{user_name}@{organization_name}.org",
                    name=user_name,
                    organization_id=organization_id,
                    roles={role},
                ),
                token=user_invitation.token,
            )
        )
        tgt_user.name = user_name
        # Verify if the right role(s) were assigned
        if tgt_user.roles != {role}:
            raise ValueError(
                f"User {tgt_user.name} has incorrect roles {tgt_user.roles}, expected {role}"
            )
        # Verify against user invitation constraints
        user_invitation_constraints: model.UserInvitationConstraints = self.handle(
            self.retrieve_invite_user_constraints_command_class(
                user=user,
            )
        )
        if tgt_user.organization_id not in user_invitation_constraints.organization_ids:
            raise ValueError("User invitation constraints not met for organization_id")
        if not tgt_user.roles.issubset(user_invitation_constraints.roles):
            raise ValueError("User invitation constraints not met for roles")
        # Verify no invitations remain for the user
        remaining_invitations: list[model.UserInvitation] = self.handle(
            self.user_invitation_crud_command_class(
                user=root_user,
                operation=CrudOperation.READ_ALL,
                obj_ids=None,
            )
        )
        if any(x.key == tgt_user.key for x in remaining_invitations):
            raise ValueError(f"Some user invitations remaining for key {tgt_user.key}")
        retval: model.User = self._set_obj(tgt_user)  # type:ignore[assignment]
        return retval

    def create_data_collection(
        self,
        user_or_str: str | model.User,
        name: str,
    ) -> model.DataCollection:
        user: model.User = self._get_obj(
            self.user_class, user_or_str
        )  # type:ignore[assignment]
        data_collection = self.handle(
            command.DataCollectionCrudCommand(
                user=user,
                operation=CrudOperation.CREATE_ONE,
                objs=model.DataCollection(
                    name=name,
                ),
            )
        )
        return self._set_obj(data_collection)  # type:ignore[return-value]

    def create_org_admin_policy(
        self,
        user_or_str: str | model.User,
        tgt_user_or_str: str | model.User,
        organization_or_str: str | model.Organization,
        is_active: bool = True,
    ) -> model.OrganizationAdminPolicy:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        tgt_user: model.User = self._get_obj(
            self.user_class, tgt_user_or_str
        )  # type:ignore[assignment]
        organization: model.Organization = self._get_obj(
            model.Organization, organization_or_str
        )  # type:ignore[assignment]
        assert organization.id is not None
        organization_admin_policy: model.OrganizationAdminPolicy = self.app.handle(
            self.organization_admin_policy_crud_command_class(
                user=user,
                operation=CrudOperation.CREATE_ONE,
                objs=self.organization_admin_policy_class(
                    organization_id=organization.id,
                    user_id=tgt_user.id,
                    is_active=is_active,
                ),
            )
        )
        return self._set_obj(organization_admin_policy)  # type:ignore[return-value]

    def read_all_user_invitations(
        self, user_or_str: str | model.User
    ) -> list[model.UserInvitation]:
        user: model.User = self._get_obj(
            self.user_class, user_or_str
        )  # type:ignore[assignment]
        invitations: list[model.UserInvitation] = self.handle(
            self.user_invitation_crud_command_class(
                user=user,
                operation=CrudOperation.READ_ALL,
            )
        )
        return invitations

    def read_organization_admin_name_emails(
        self, user_or_str: str | model.User
    ) -> list[model.UserNameEmail]:
        user: model.User = self._get_obj(
            self.user_class, user_or_str
        )  # type:ignore[assignment]
        user_name_emails: list[model.UserNameEmail] = self.app.handle(
            self.retrieve_organization_admin_name_emails_command_class(user=user)
        )
        return user_name_emails

    def read_all_users(self) -> list[model.User]:
        root_user = self.get_root_user()
        retval: list[model.User] = self.app.handle(
            self.user_crud_command_class(
                user=root_user,
                operation=CrudOperation.READ_ALL,
            )
        )
        return retval

    def read_users_by_role(self, role: Enum) -> list[model.User]:
        users = self.read_all_users()
        return [x for x in users if role in x.roles]

    def update_user(
        self,
        user_or_str: str | model.User,
        tgt_user_or_str: str | model.User,
        is_active: bool | None = None,
        roles: set[str] | None = None,
        organization_or_str: str | None = None,
        set_dummy_organization: bool = False,
    ) -> model.User:
        user: model.User = self._get_obj(
            self.user_class, user_or_str
        )  # type:ignore[assignment]
        tgt_user: model.User = self._get_obj(
            self.user_class, tgt_user_or_str, copy=True
        )  # type:ignore[assignment]
        if not organization_or_str:
            if set_dummy_organization:
                organization_id = self.generate_id()
            else:
                organization_id = None
        else:
            if set_dummy_organization:
                raise ValueError("Organization given and set_dummy_organization True")
            organization: model.Organization = self._get_obj(
                model.Organization, organization_or_str
            )  # type:ignore[assignment]
            organization_id = organization.id
        has_updates = False
        if is_active is not None and tgt_user.is_active != is_active:
            has_updates = True
            tgt_user.is_active = is_active
        if roles is not None and tgt_user.roles != roles:
            has_updates = True
            tgt_user.roles = roles
        if organization_id is not None and tgt_user.organization_id != organization_id:
            has_updates = True
            tgt_user.organization_id = organization_id
        sleep(0.000000001)  # To avoid having same _modified_at as tgt_user
        assert tgt_user.id is not None
        updated_tgt_user = self.handle(
            self.update_user_command_class(
                user=user,
                tgt_user_id=tgt_user.id,
                is_active=tgt_user.is_active,
                roles=tgt_user.roles,
                organization_id=tgt_user.organization_id,
            )
        )
        updated_tgt_user.name = tgt_user.name
        TestClient._verify_updated_obj(
            tgt_user, updated_tgt_user, user.id, verify_modified=has_updates
        )
        return self._set_obj(updated_tgt_user, update=True)  # type:ignore[return-value]

    def get_root_user(self, user_key: str | None = None) -> model.User:
        if user_key is None:
            user_key = self.cfg["service"]["auth"]["props"]["root"]["user"]["key"]
        user: model.User = self.app.user_manager.retrieve_user_by_key(user_key)  # type: ignore[assignment]
        return user

    def get_org_ids_for_org_admin(
        self,
        user_or_str: str | model.User,
        include_self: bool = False,
        on_no_admin: str = "raise",
    ) -> list[model.Organization]:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        org_admin_policies: list[model.OrganizationAdminPolicy] = [  # type: ignore[assignment]
            x
            for x in self.db[self.organization_admin_policy_class].values()
            if x.user_id == user.id
        ]
        if not org_admin_policies:
            if on_no_admin == "raise":
                raise ValueError(f"User {user.name} is not an organization admin")
            elif on_no_admin == "return":
                return []
        organization_ids = {x.organization_id for x in org_admin_policies}
        if include_self:
            organization_ids.add(user.organization_id)
        return organization_ids

    def get_own_org_admin_users(
        self,
        user_or_str: str | model.User,
        include_self: bool = False,
    ) -> list[model.Organization]:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        org_admin_policies: list[model.OrganizationAdminPolicy] = [
            x
            for x in self.db[self.organization_admin_policy_class].values()
            if x.organization_id == user.organization_id and x.is_active
        ]
        user_ids = {x.user_id for x in org_admin_policies}
        if include_self:
            user_ids.add(user.id)
        return [
            x
            for x in self.db[self.user_class].values()
            if x.id in user_ids and x.is_active
        ]

    def get_users_for_org(
        self,
        user_or_str: str | model.User,
        on_no_admin: str = "raise",
    ) -> list[model.User]:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        return [
            x
            for x in self.db[self.user_class].values()
            if x.organization_id == user.organization_id
        ]

    def get_users_for_org_admin(
        self,
        user_or_str: str | model.User,
        include_self: bool = False,
        include_other_org_admins: bool = False,
        on_no_admin: str = "raise",
    ) -> list[model.User]:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        org_admin_policies: list[model.OrganizationAdminPolicy] = (
            [  # type:ignore[assignment]
                x
                for x in self.db[self.organization_admin_policy_class].values()
                if x.user_id == user.id and x.is_active
            ]
        )
        if not org_admin_policies:
            if on_no_admin == "raise":
                raise ValueError(f"User {user.name} is not an organization admin")
            elif on_no_admin == "return":
                return []
        organization_ids = {x.organization_id for x in org_admin_policies}
        user_ids = {user.id} if include_self else set()
        if include_other_org_admins:
            other_org_admin_policies: list[model.OrganizationAdminPolicy] = (
                [  # type:ignore[assignment]
                    x
                    for x in self.db[self.organization_admin_policy_class].values()
                    if x.organization_id in organization_ids and x.is_active
                ]
            )
            user_ids.update({x.user_id for x in other_org_admin_policies})
        tgt_users: list[model.User] = list(self.db[self.user_class].values())
        return [
            x
            for x in tgt_users
            if x.id in user_ids or x.organization_id in organization_ids
        ]

    def is_sub_role(
        self,
        sub_role: str,
        role: str,
        allow_equal: bool = False,
    ) -> bool:
        """
        Check if sub_role is indeed a sub-role of role based on the permissions
        they each have. Set allow_equal=True to allow sub_role to be equal to role.
        """
        permissions = self.role_permissions_map[role]
        sub_permissions = self.role_permissions_map[sub_role]
        if allow_equal:
            return sub_permissions.issubset(permissions[role])
        return len(permissions) != len(sub_permissions) and sub_permissions.issubset(
            permissions
        )

    def read_all(
        self,
        user_or_key: model.User | str,
        model_class: type[model.Model],
        cascade: bool = False,
    ) -> list[model.Model]:
        user: model.User = self._get_obj(self.user_class, user_or_key)  # type: ignore[assignment]
        retval: list[model.Model] = self.handle(
            self.app.domain.get_crud_command_for_model(model_class)(
                user=user,
                operation=CrudOperation.READ_ALL,
                props={"cascade_read": cascade},
            ),
            use_endpoint=False,
        )
        return retval

    def read_some(
        self,
        user_or_key: model.User | str,
        model_class: type[model.Model],
        obj_ids: list[UUID] | set[UUID],
        cascade: bool = False,
    ) -> list[model.Model]:
        user: model.User = self._get_obj(self.user_class, user_or_key)  # type: ignore[assignment]
        retval: list[model.Model] = self.handle(
            self.app.domain.get_crud_command_for_model(model_class)(
                user=user,
                operation=CrudOperation.READ_SOME,
                obj_ids=(
                    list(obj_ids)
                    if isinstance(obj_ids, set)
                    else obj_ids  # type:ignore[arg-type]
                ),
                props={"cascade_read": cascade},
            ),
            use_endpoint=False,
        )
        return retval

    def read_some_by_property(
        self,
        user_or_key: model.User | str,
        model_class: type[model.Model],
        name: str,
        value: Any,
        cascade: bool = False,
    ) -> list[model.Model]:
        objs = self.read_all(user_or_key, model_class, cascade=cascade)
        return [x for x in objs if getattr(x, name) == value]

    def read_one_by_property(
        self,
        user_or_key: model.User | str,
        model_class: type[model.Model],
        name: str,
        value: Any,
        cascade: bool = False,
    ) -> model.Model:
        objs = self.read_some_by_property(
            user_or_key, model_class, name, value, cascade=cascade
        )
        if len(objs) == 0:
            raise ValueError(f"{model_class} with {name}='{value}' not found")
        if len(objs) > 1:
            raise ValueError(f"Multiple {model_class} with {name}='{value}' found")
        return objs[0]

    def update_object(
        self,
        user_or_key: str | model.User,
        model_class: type[model.Model],
        obj_or_key: model.Model | str,
        props: dict[str, Any | None],
        set_dummy_link: dict[str, bool] | bool = False,
        exclude_none: bool = True,
    ) -> model.Model:
        user: model.User = self._get_obj(self.user_class, user_or_key)  # type: ignore[assignment]
        obj: model.Model = self._get_obj(
            model_class, obj_or_key, copy=True
        )  # type:ignore[assignment]
        self._update_object_properties(
            obj, props, set_dummy_link, exclude_none=exclude_none
        )
        sleep(0.000000001)
        updated_obj = self.handle(
            self.app.domain.get_crud_command_for_model(model_class)(
                user=user,
                operation=CrudOperation.UPDATE_ONE,
                objs=obj,
            )
        )
        assert user.id
        TestClient._verify_updated_obj(obj, updated_obj, user.id)
        return self._set_obj(updated_obj, update=True)

    def delete_object(
        self,
        user_or_key: str | model.User,
        model_class: type[model.Model],
        obj_or_key: model.Model | str | tuple[UUID, UUID],
        retry_obj: tuple[UUID, UUID] | None = None,
    ) -> UUID:
        user: model.User = self._get_obj(self.user_class, user_or_key)  # type: ignore[assignment]
        obj: model.Model = self._get_obj(model_class, obj_or_key, copy=True)  # type: ignore[assignment]

        if not obj and retry_obj:
            obj = self._get_obj(model_class, retry_obj, copy=True)  # type: ignore[assignment]

        deleted_obj_id: UUID = self.handle(
            self.app.domain.get_crud_command_for_model(model_class)(
                user=user,
                operation=CrudOperation.DELETE_ONE,
                obj_ids=obj.id,
            )
        )
        # verify if deleted
        # is_existing_obj = self.app.handle(
        #     self.app.domain.get_crud_command_for_model(model_class)(
        #         user=user,
        #         operation=CrudOperation.EXISTS_ONE,
        #         obj_ids=deleted_obj_id,
        #     )
        # )
        # if is_existing_obj:
        #     raise ValueError(f"Object {deleted_obj_id} not deleted")
        self._delete_obj(model_class, deleted_obj_id)
        return deleted_obj_id

    def verify_read_all(
        self,
        user_or_str: model.User | str,
        model_class: type[model.Model],
        expected_ids: set[UUID] | list[model.Model],
    ) -> None:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        objs = self.handle(
            self.app.domain.get_crud_command_for_model(model_class)(
                user=user, operation=CrudOperation.READ_ALL
            )
        )
        actual_ids = {x.id for x in objs}
        if not isinstance(expected_ids, set):
            expected_ids = {x.id for x in expected_ids if x.id is not None}
        if actual_ids != expected_ids:
            extra_ids = actual_ids - expected_ids
            missing_ids = expected_ids - actual_ids
            extra_names = [
                self._get_key_for_obj(
                    self._get_obj(model_class, x)  # type:ignore[arg-type]
                )
                for x in extra_ids
            ]
            missing_names = [
                self._get_key_for_obj(
                    self._get_obj(model_class, x)  # type:ignore[arg-type]
                )
                for x in missing_ids
            ]
            raise ValueError(
                f"Difference in read all. Extra: {extra_names}. Missing: {missing_names}. User: {user.name}. Model: {model_class}"
            )

    def verify_user_has_role(
        self, user_or_str: str | model.User, role: Enum, exclusive: bool = True
    ) -> bool:
        user: model.User = self._get_obj(
            self.user_class, user_or_str
        )  # type:ignore[assignment]
        roles = user.roles
        if exclusive:
            return role in roles and len(roles) == 1
        return role in roles

    def generate_id(self) -> UUID:
        # Type cast needed because app.generate_id() returns Hashable
        return cast(UUID, self.app.generate_id())

    def print_organizations(self) -> None:
        organizations: list[model.Organization] = self.read_all(
            self.get_root_user(), model.Organization, cascade=True
        )  # type:ignore[assignment]
        print("\nOrganizations:")
        for x in sorted(organizations, key=lambda x: x.name):
            print(f"{x.name} ({x.id})")

    def print_data_collections(self) -> None:
        data_collections: list[model.DataCollection] = self.read_all(
            self.get_root_user(), model.DataCollection, cascade=True
        )  # type:ignore[assignment]
        print("\nDataCollections:")
        for x in sorted(data_collections, key=lambda x: x.name):
            print(f"{x.name} ({x.id})")

    def print_users(self) -> None:
        root_user: model.User = self.get_root_user()
        users: list[model.User] = self.read_all(
            root_user, self.user_class
        )  # type:ignore[assignment]
        organizations: dict[UUID, model.Organization] = {
            x.id: x  # type:ignore[misc]
            for x in self.read_all(root_user, model.Organization)
        }
        print("\nUsers:")
        for x in sorted(
            users, key=lambda x: (organizations[x.organization_id].name, x.key)
        ):
            print(
                f"{organizations[x.organization_id].name} / {x.key}: "
                + ", ".join(
                    [z for z in sorted(self.rev_role_map[y].name for y in x.roles)]
                )
                + f" ({x.id})"
            )

    def print_user_permissions(self, user_or_str: str | model.User) -> None:
        user: model.User = self._get_obj(self.user_class, user_or_str)  # type: ignore[assignment]
        user_permissions = self.app.user_manager.retrieve_user_permissions(user)
        command_permissions = map_paired_elements(
            ((x.command_name, x.permission_type) for x in user_permissions), as_set=True
        )
        print(
            f"\nPermissions for user {user.name} (n_commands={len(command_permissions)}):"
        )
        for x in sorted(list(user_permissions), key=lambda x: x.sort_key):  # type: ignore[arg-type,return-value]
            print(f"{x}")

    def print_org_admin_policies(self) -> None:
        org_admin_policies: list[model.OrganizationAdminPolicy] = self.read_all(  # type: ignore[assignment]
            "root1_1", self.organization_admin_policy_class, cascade=True
        )
        print("\nOrganizationAdminPolicies:")
        for x in sorted(
            org_admin_policies,
            key=lambda x: (x.organization.name, x.user.name),  # type:ignore[union-attr]
        ):
            print(
                f"{x.organization.name}: user={x.user.name} (is_active={x.is_active}) ({x.id})"  # type:ignore[union-attr]
            )

    def _get_key_for_obj(self, obj: model.Model) -> Any:
        key_fields = self.MODEL_KEY_MAP[obj.__class__]
        if isinstance(key_fields, str):
            return getattr(obj, key_fields)
        return tuple(getattr(obj, x) for x in key_fields)

    def _update_object_properties(
        self,
        obj: model.Model,
        props: dict[str, Any | None],
        set_dummy_link: dict[str, bool] | bool = False,
        exclude_none: bool = True,
    ) -> None:
        """
        Helper function for update methods. All the (field_name, value) pairs in props
        are set as attributes of obj. If the field_name is a relationship field, the value
        is set as the id of the linked object. If set_dummy_link is provided for a
        relationship field and no real linked obj is provided, a dummy id is put instead.
        If exclude_none is True, fields or link fields with value None are not set.
        """
        # Parse input
        model_class = obj.__class__
        assert model_class.ENTITY
        id_field_name = model_class.ENTITY.id_field_name
        assert id_field_name
        link_map: dict[str, tuple[str, type[model.Model]]] = {
            x.relationship_field_name: (
                x.link_field_name,
                x.link_model_class,
            )  # type:ignore[misc]
            for x in model_class.ENTITY.links.values()
            if x.relationship_field_name
        }
        default_set_dummy_link = False
        if isinstance(set_dummy_link, bool):
            default_set_dummy_link = set_dummy_link
            set_dummy_link = {}

        # Set value fields and any links
        for field_name, value in props.items():
            if field_name in link_map:
                field_name, link_model_class = link_map[field_name]
                if not value:
                    if set_dummy_link.get(field_name, default_set_dummy_link):
                        value = self.generate_id()
                    else:
                        value = None
                else:
                    if set_dummy_link.get(field_name, default_set_dummy_link):
                        raise ValueError(
                            f"{model_class.__name__} given and set dummy link True"
                        )
                    value = getattr(
                        self._get_obj(link_model_class, value), id_field_name
                    )
            if exclude_none and value is None:
                continue
            setattr(obj, field_name, value)

    def _get_obj_key(
        self,
        table: dict,
        model_class: type[model.Model],
        obj: (
            str
            | UUID
            | model.Model
            | list[str | UUID | model.Model]
            | tuple[UUID, UUID]
        ),
        on_missing: str,
    ) -> tuple[UUID, UUID] | UUID | None:
        key_fields = self.MODEL_KEY_MAP[model_class]
        assert model_class.ENTITY
        get_obj_id = model_class.ENTITY.get_obj_id
        if not isinstance(key_fields, tuple):
            key_fields = (key_fields,) if len(key_fields) > 1 else key_fields
        if isinstance(obj, str) or isinstance(obj, datetime.datetime):
            key = (obj,)
        elif isinstance(obj, UUID):
            key = [
                x for x, y in table.items() if get_obj_id(y) == obj
            ]  # type:ignore[assignment]
            if key:
                pass
            elif on_missing == "raise":
                raise ValueError(f"{model_class.__name__} {obj} not found")
            elif on_missing == "return_none":
                return None
            else:
                raise NotImplementedError()
        elif isinstance(obj, model.Model):
            key = tuple(getattr(obj, x) for x in key_fields)
        elif isinstance(obj, tuple):
            key = obj  # type:ignore[assignment]
        else:
            raise ValueError(f"Invalid object: {obj}")
        key = key if len(key) > 1 else key[0]  # type:ignore[assignment]
        return key  # type:ignore[return-value]

    def _get_obj(
        self,
        model_class: type[model.Model],
        obj: (
            str
            | UUID
            | model.Model
            | list[str | UUID | model.Model]
            | tuple[UUID, UUID]
        ),
        copy: bool = False,
        on_missing: str = "raise",
    ) -> model.Model | list[model.Model] | None:
        if model_class not in self.db:
            self.db[model_class] = {}
        if isinstance(obj, list):
            return [self._get_obj(model_class, x) for x in obj]  # type: ignore[misc]
        table = self.db[model_class]
        key = self._get_obj_key(table, model_class, obj, on_missing)
        if key not in table:
            if on_missing == "raise":
                raise ValueError(f"{model_class.__name__} {obj} not found")
            elif on_missing == "return_none":
                return None
            else:
                raise NotImplementedError()
        return table[key] if not copy else table[key].model_copy()

    def _set_obj(self, obj: model.Model, update: bool = False) -> model.Model:
        model_class = type(obj)
        if model_class not in self.db:
            self.db[model_class] = {}
        table = self.db[model_class]
        key_fields = self.MODEL_KEY_MAP[model_class]
        if not isinstance(key_fields, tuple):
            key_fields = (key_fields,) if len(key_fields) > 1 else key_fields
        model_name = model_class.__name__
        key = tuple(getattr(obj, x) for x in key_fields)
        key = (
            key if len(key) > 1 else key[0]  # pyright: ignore[reportGeneralTypeIssues]
        )
        if key in table:
            if update:
                table[key] = obj
            else:
                raise ValueError(f"{model_name} {obj} already exists")
        else:
            table[key] = obj
        return obj

    def _delete_obj(self, model_class: type[model.Model], obj_id: UUID) -> model.Model:
        if model_class not in self.db:
            self.db[model_class] = {}
        table = self.db[model_class]
        key = [x for x, y in table.items() if y.id == obj_id]
        if key:
            key = key[0]  # type:ignore[assignment]
        else:
            raise ValueError(f"{model_class} {obj_id} not found")
        if key not in table:  # type:ignore[comparison-overlap]
            raise ValueError(f"{model_class} {obj_id} not found")
        obj = table[key]  # type:ignore[index]
        del table[key]  # type:ignore[arg-type]
        return obj

    @staticmethod
    def _set_log_level(app_cfg: BaseAppCfg, log_level: int) -> None:
        set_log_level(app_cfg.app_name.lower(), log_level)

    @staticmethod
    def _verify_updated_obj(
        in_obj: model.Model, out_obj: model.Model, user_id: UUID, **kwargs: Any
    ) -> None:
        # TODO: verifying modified_by and modified_at is no longer possible here as the
        # persistence metadata no longer exists in the object. This should instead
        # be tested through unit tests on the repository in question.
        # verify_modified = kwargs.get("verify_modified", True)
        # if verify_modified and out_obj._modified_by != user_id:
        #     raise ValueError(f"_modified_by not updated: {out_obj._modified_by}")
        # if verify_modified and out_obj._modified_at <= in_obj._modified_at:
        #     raise ValueError(f"modified_at not updated: {out_obj._modified_at}")
        # if (
        #     out_obj.model_copy(
        #         update={
        #             "_modified_by": in_obj._modified_by,
        #             "_modified_at": in_obj._modified_at,
        #         }
        #     )
        #     != in_obj
        # ):
        #     raise ValueError(f"Object not updated: {in_obj}, {out_obj}")
        if out_obj != in_obj:
            raise ValueError(f"Object not updated: {in_obj}, {out_obj}")



from gen_epix.commondb.domain.enum import Role as CommonRole
from gen_epix.commondb.domain.policy import RoleGenerator as CommonRoleGenerator
from gen_epix.fastapp import PermissionTypeSet
from gen_epix.fastapp.services.rbac import BaseRbacService
from gen_epix.seqdb.domain import command
from gen_epix.seqdb.domain.enum import Role


class RoleGenerator(CommonRoleGenerator):

    COMMON_ROLE_ENUM_MAP = {x: Role[x.name] for x in CommonRole}

    EXTRA_ROLE_SET_MAP = {}

    COMMON_ROLE_PERMISSION_SETS = (
        CommonRoleGenerator.map_from_common_role_permission_sets(
            COMMON_ROLE_ENUM_MAP, command.COMMON_COMMAND_MAP
        )
    )

    ROLE_PERMISSION_SETS = {
        Role.APP_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.APP_ADMIN] | set(),
        Role.REFDATA_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.REFDATA_ADMIN]
        | {
            # seq.metadata CRUD commands
            (command.AlignmentProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.AssemblyProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.AstProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpCrudCommand, PermissionTypeSet.CRU),
            (command.SnpDetectionProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpSetCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.LibraryPrepProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.LocusCrudCommand, PermissionTypeSet.CRU),
            (command.LocusDetectionProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.LocusSetCrudCommand, PermissionTypeSet.CRU),
            (command.LocusSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.RefAlleleCrudCommand, PermissionTypeSet.CRU),
            (command.PcrProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSeqCrudCommand, PermissionTypeSet.CRU),
            (command.SeqCategoryCrudCommand, PermissionTypeSet.CRU),
            (command.SeqCategorySetCrudCommand, PermissionTypeSet.CRU),
            (command.SeqClassificationProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.SeqDistanceProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.SubtypingSchemeCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonLocusLinkCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonSetCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.CRU),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonomyProtocolCrudCommand, PermissionTypeSet.CRU),
        },
        Role.ORG_ADMIN: COMMON_ROLE_PERMISSION_SETS[Role.ORG_ADMIN] | set(),
        Role.ORG_USER: COMMON_ROLE_PERMISSION_SETS[Role.ORG_USER]
        | {
            # seq.persistable CRUD commands
            (command.AlleleCrudCommand, PermissionTypeSet.CRUD),
            (command.AlleleAlignmentCrudCommand, PermissionTypeSet.CRUD),
            (command.SampleCrudCommand, PermissionTypeSet.CRUD),
            (command.ReadSetCrudCommand, PermissionTypeSet.CRUD),
            (command.FileCrudCommand, PermissionTypeSet.CRUD),
            (command.SeqCrudCommand, PermissionTypeSet.CRUD),
            (command.RawSeqCrudCommand, PermissionTypeSet.CRUD),
            (command.AlleleProfileCrudCommand, PermissionTypeSet.CRUD),
            (command.SeqAlignmentCrudCommand, PermissionTypeSet.CRUD),
            (command.SeqTaxonomyCrudCommand, PermissionTypeSet.CRUD),
            (command.PcrMeasurementCrudCommand, PermissionTypeSet.CRUD),
            (command.AstMeasurementCrudCommand, PermissionTypeSet.CRUD),
            (command.AstPredictionCrudCommand, PermissionTypeSet.CRUD),
            (command.SeqDistanceCrudCommand, PermissionTypeSet.CRUD),
            (command.SeqClassificationCrudCommand, PermissionTypeSet.CRUD),
            (command.SnpProfileCrudCommand, PermissionTypeSet.CRUD),
            # seq.metadata CRUD commands
            (command.AlignmentProtocolCrudCommand, PermissionTypeSet.R),
            (command.AssemblyProtocolCrudCommand, PermissionTypeSet.R),
            (command.AstProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSnpCrudCommand, PermissionTypeSet.R),
            (command.SnpDetectionProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSnpSetCrudCommand, PermissionTypeSet.R),
            (command.RefSnpSetMemberCrudCommand, PermissionTypeSet.R),
            (command.LibraryPrepProtocolCrudCommand, PermissionTypeSet.R),
            (command.LocusCrudCommand, PermissionTypeSet.R),
            (command.LocusDetectionProtocolCrudCommand, PermissionTypeSet.R),
            (command.LocusSetCrudCommand, PermissionTypeSet.R),
            (command.LocusSetMemberCrudCommand, PermissionTypeSet.R),
            (command.PcrProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSeqCrudCommand, PermissionTypeSet.R),
            (command.SeqCategoryCrudCommand, PermissionTypeSet.R),
            (command.SeqCategorySetCrudCommand, PermissionTypeSet.R),
            (command.SeqClassificationProtocolCrudCommand, PermissionTypeSet.R),
            (command.SeqDistanceProtocolCrudCommand, PermissionTypeSet.R),
            (command.SubtypingSchemeCrudCommand, PermissionTypeSet.R),
            (command.TaxonCrudCommand, PermissionTypeSet.R),
            (command.TaxonLocusLinkCrudCommand, PermissionTypeSet.R),
            (command.TaxonSetCrudCommand, PermissionTypeSet.R),
            (command.TaxonSetMemberCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.R),
            (command.TaxonomyProtocolCrudCommand, PermissionTypeSet.R),
            # seq non-CRUD commands
            (command.RetrievePhylogeneticTreeCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteAlleleProfileCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSnpProfileCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteContigCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSampleCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSeqCommand, PermissionTypeSet.E),
            (command.RetrieveSeqFastaCommand, PermissionTypeSet.E),
        },
        Role.GUEST: COMMON_ROLE_PERMISSION_SETS[Role.GUEST] | set(),
    }

    ROLE_HIERARCHY = CommonRoleGenerator.map_from_common_role_hierarchy(
        COMMON_ROLE_ENUM_MAP
    )

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSION_SETS
    )



# pylint: disable=useless-import-alias
from gen_epix.commondb.api.auth import create_auth_endpoints as create_auth_endpoints
from gen_epix.commondb.api.organization import (
    UpdateDataCollectionSetDataCollectionRequestBody as UpdateDataCollectionSetDataCollectionRequestBody,
)
from gen_epix.commondb.api.organization import (
    UpdateOrganizationSetOrganizationRequestBody as UpdateOrganizationSetOrganizationRequestBody,
)
from gen_epix.commondb.api.organization import (
    UpdateUserOwnOrganizationRequestBody as UpdateUserOwnOrganizationRequestBody,
)
from gen_epix.commondb.api.organization import (
    UpdateUserRequestBody as UpdateUserRequestBody,
)
from gen_epix.commondb.api.organization import (
    UserInvitationRequestBody as UserInvitationRequestBody,
)
from gen_epix.commondb.api.organization import (
    create_organization_endpoints as create_organization_endpoints,
)
from gen_epix.commondb.api.rbac import create_rbac_endpoints as create_rbac_endpoints
from gen_epix.commondb.api.system import HealthReponseBody as HealthReponseBody
from gen_epix.commondb.api.system import HealthStatus as HealthStatus
from gen_epix.commondb.api.system import LogItem as LogItem
from gen_epix.commondb.api.system import LogRequestBody as LogRequestBody
from gen_epix.commondb.api.system import (
    create_system_endpoints as create_system_endpoints,
)



import logging
from collections.abc import Callable

from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware

from gen_epix.fastapp.app import App
from gen_epix.fastapp.exc import AuthException


class HandleAuthExceptionMiddleware(BaseHTTPMiddleware):
    def __init__(
        # Mandatory parameters put as kwargs to comply with the signature of
        # BaseHTTPMiddleware
        self,
        app: FastAPI,
        fast_app: App = None,  # type: ignore[assignment]
        logger: logging.Logger | None = None,
    ):
        super().__init__(app)
        self._fast_app = fast_app
        self._logger = logger or fast_app.logger

    def _log_exception(self, exception: Exception) -> None:
        if self._logger:
            self._logger.warning(
                self._fast_app.create_log_message(
                    "e4cf7b23", None, exception=exception  # type: ignore[arg-type]
                ),
            )

    async def dispatch(self, request: Request, call_next: Callable) -> Response:
        try:
            response: Response = await call_next(request)
            return response
        except* AuthException as exception_group:
            if self._logger:
                for exception in exception_group.exceptions:
                    self._log_exception(exception)
        # TODO: check if other domain exceptions need to be caught here
        return JSONResponse(None, status_code=401)



import abc
from collections.abc import Iterable
from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository import BaseCaseRepository
from gen_epix.fastapp import BaseService


class BaseCaseService(BaseService):
    SERVICE_TYPE = ServiceType.CASE

    NO_ABAC_COMMAND_CLASSES: set[type[command.Command]] = {
        command.TreeAlgorithmClassCrudCommand,
        command.TreeAlgorithmCrudCommand,
        command.GeneticDistanceProtocolCrudCommand,
        command.DimCrudCommand,
        command.ColCrudCommand,
        command.CaseTypeSetCategoryCrudCommand,
        command.CaseSetCategoryCrudCommand,
        command.CaseSetStatusCrudCommand,
    }
    ABAC_METADATA_COMMAND_CLASSES: set[type[command.Command]] = {
        command.CaseTypeCrudCommand,
        command.CaseTypeSetMemberCrudCommand,
        command.CaseTypeSetCrudCommand,
        command.CaseTypeColCrudCommand,
        command.CaseTypeColSetMemberCrudCommand,
        command.CaseTypeColSetCrudCommand,
    }
    ABAC_DATA_COMMAND_CLASSES: set[type[command.Command]] = {
        command.CaseCrudCommand,
        command.CaseSetCrudCommand,
        command.CaseSetMemberCrudCommand,
        command.CaseDataCollectionLinkCrudCommand,
        command.CaseSetDataCollectionLinkCrudCommand,
        command.ValidateCasesCommand,
    }
    CASCADE_DELETE_MODEL_CLASSES: dict[
        type[model.Model], tuple[type[model.Model], ...]
    ] = {
        model.CaseTypeSet: (model.CaseTypeSetMember,),
        model.CaseType: (model.CaseTypeSetMember,),
        model.CaseTypeColSet: (model.CaseTypeColSetMember,),
        model.CaseTypeCol: (model.CaseTypeColSetMember,),
        model.CaseSet: (
            model.CaseSetDataCollectionLink,
            model.CaseSetMember,
        ),
        model.Case: (
            model.CaseDataCollectionLink,
            model.CaseSetMember,
        ),
    }

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseCaseRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        self.register_default_crud_handlers()
        f(command.ValidateCasesCommand, self.validate_cases)
        f(command.CreateCasesCommand, self.create_cases)
        f(command.CreateCaseSetCommand, self.create_case_set)
        f(command.RetrieveCompleteCaseTypeCommand, self.retrieve_complete_case_type)
        f(command.RetrieveCaseTypeStatsCommand, self.retrieve_case_type_stats)
        f(command.RetrieveCaseSetStatsCommand, self.retrieve_case_set_stats)
        f(command.RetrieveCasesByQueryCommand, self.retrieve_cases_by_query)
        f(command.RetrieveCasesByIdCommand, self.retrieve_cases_by_id)
        f(command.RetrieveCaseRightsCommand, self.retrieve_case_or_set_rights)
        f(command.RetrieveCaseSetRightsCommand, self.retrieve_case_or_set_rights)
        f(
            command.RetrieveGeneticSequenceByCaseCommand,
            self.retrieve_genetic_sequence_by_case,
        )
        f(
            command.RetrievePhylogeneticTreeByCasesCommand,
            self.retrieve_phylogenetic_tree,
        )
        f(
            command.RetrieveGeneticSequenceByCaseCommand,
            self.retrieve_genetic_sequence_by_case,
        )
        f(
            command.RetrieveGeneticSequenceFastaByCaseCommand,
            self.retrieve_genetic_sequence_fasta_by_case,
        )
        f(command.CreateReadSetsForCasesCommand, self.create_reads_sets_for_cases)
        f(command.CreateFileForReadSetCommand, self.create_file_for_read_set)
        f(command.CreateSeqsForCasesCommand, self.create_seqs_for_cases)
        f(command.CreateFileForSeqCommand, self.create_file_for_seq)
        f(
            command.RetrieveLibraryPrepProtocolsCommand,
            self.retrieve_library_prep_protocols,
        )
        f(
            command.RetrieveAssemblyProtocolsCommand,
            self.retrieve_assembly_protocols,
        )

    @abc.abstractmethod
    def validate_cases(
        self, cmd: command.ValidateCasesCommand
    ) -> model.CaseValidationReport:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_cases(self, cmd: command.CreateCasesCommand) -> list[model.Case] | None:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_case_set(
        self, cmd: command.CreateCaseSetCommand
    ) -> model.CaseSet | None:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_complete_case_type(
        self,
        cmd: command.RetrieveCompleteCaseTypeCommand,
    ) -> model.CompleteCaseType:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_set_stats(
        self,
        cmd: command.RetrieveCaseSetStatsCommand,
    ) -> list[model.CaseSetStat]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_type_stats(
        self,
        cmd: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_cases_by_query(
        self, cmd: command.RetrieveCasesByQueryCommand
    ) -> list[UUID]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_cases_by_id(
        self, cmd: command.RetrieveCasesByIdCommand
    ) -> list[model.Case]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_or_set_rights(
        self,
        cmd: command.RetrieveCaseRightsCommand | command.RetrieveCaseSetRightsCommand,
    ) -> list[model.CaseRights] | list[model.CaseSetRights]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_phylogenetic_tree(
        self,
        cmd: command.RetrievePhylogeneticTreeByCasesCommand,
    ) -> model.PhylogeneticTree:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_genetic_sequence_by_case(
        self,
        cmd: command.RetrieveGeneticSequenceByCaseCommand,
    ) -> list[model.GeneticSequence]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_genetic_sequence_fasta_by_case(
        self,
        cmd: command.RetrieveGeneticSequenceFastaByCaseCommand,
    ) -> Iterable[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_reads_sets_for_cases(
        self,
        cmd: command.CreateReadSetsForCasesCommand,
    ) -> list[model.ReadSet]:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_file_for_read_set(
        self,
        cmd: command.CreateFileForReadSetCommand,
    ) -> UUID:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_seqs_for_cases(
        self,
        cmd: command.CreateSeqsForCasesCommand,
    ) -> list[model.Seq]:
        raise NotImplementedError()

    @abc.abstractmethod
    def create_file_for_seq(
        self,
        cmd: command.CreateFileForSeqCommand,
    ) -> UUID:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_library_prep_protocols(
        self,
        cmd: command.RetrieveLibraryPrepProtocolsCommand,
    ) -> list[model.LibraryPrepProtocol]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_assembly_protocols(
        self,
        cmd: command.RetrieveAssemblyProtocolsCommand,
    ) -> list[model.AssemblyProtocol]:
        raise NotImplementedError()



from typing import Any

from gen_epix.fastapp.app import App
from gen_epix.fastapp.model import Command
from gen_epix.fastapp.service import BaseService


class BaseRemoteService(BaseService):
    def __init__(self, app: App, remote_app: App, use_endpoints: bool = True) -> None:
        super().__init__(app)
        self._remote_app = remote_app
        self._use_endpoints = use_endpoints

    @property
    def remote_app(self) -> App:
        return self._remote_app

    @property
    def use_endpoints(self) -> bool:
        return self._use_endpoints

    @use_endpoints.setter
    def use_endpoints(self, value: bool) -> None:
        self._use_endpoints = value

    def handle(
        self,
        cmd: Command,
    ) -> Any:
        if self.use_endpoints:
            raise NotImplementedError()
        else:
            return self.remote_app.handle(cmd)



from typing import ClassVar
from uuid import UUID

from pydantic import Field

from gen_epix.commondb.domain.model import Model
from gen_epix.fastapp.domain import Entity
from gen_epix.omopdb.domain.model.omop.omop import (
    DrugExposure,
    LocationHistory,
    Measurement,
    Observation,
    Person,
    Specimen,
)


class Subject(Model):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="subjects",
        persistable=False,
    )

    id: UUID | None = Field(default=None, description="The ID of the subject.")
    person: Person | None = Field(
        default=None, description="The person associated with the subject."
    )
    specimen_records: list[Specimen] = Field(
        description="The list of specimen records associated with the subject."
    )
    observation_records: list[Observation] = Field(
        description="The list of observations records associated with the subject."
    )
    measurement_records: list[Measurement] = Field(
        description="The list of measurements records associated with the subject."
    )
    drug_exposure_records: list[DrugExposure] = Field(
        description="The list of drug exposure records associated with the subject."
    )
    location_history_records: list[LocationHistory] = Field(
        description="The list of location history records associated with the subject."
    )



import datetime
from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped, declarative_mixin, declared_attr, relationship

from gen_epix.commondb.domain import DOMAIN, enum, model
from gen_epix.commondb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.commondb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: type = orm.declarative_base(name=enum.ServiceType.ORGANIZATION.value)


@declarative_mixin
class OrganizationMixin(RowMetadataMixin):
    name: Mapped[str] = create_mapped_column(DOMAIN, model.Organization, "name")
    legal_entity_code: Mapped[str] = create_mapped_column(
        DOMAIN, model.Organization, "legal_entity_code"
    )


@declarative_mixin
class UserMixin(RowMetadataMixin):
    key: Mapped[str] = create_mapped_column(DOMAIN, model.User, "key")
    email: Mapped[str | None] = create_mapped_column(DOMAIN, model.User, "email")
    name: Mapped[str | None] = create_mapped_column(DOMAIN, model.User, "name")
    is_active: Mapped[bool] = create_mapped_column(DOMAIN, model.User, "is_active")
    roles: Mapped[set[str]] = create_mapped_column(DOMAIN, model.User, "roles")

    @declared_attr
    def organization_id(cls) -> Mapped[UUID]:
        return create_mapped_column(DOMAIN, model.User, "organization_id")

    @declared_attr
    def organization(cls) -> Mapped[model.Organization]:
        return relationship("Organization", foreign_keys="User.organization_id")


@declarative_mixin
class OrganizationSetMixin(RowMetadataMixin):
    name: Mapped[str] = create_mapped_column(DOMAIN, model.OrganizationSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.OrganizationSet, "description"
    )


@declarative_mixin
class OrganizationSetMemberMixin(RowMetadataMixin):
    @declared_attr
    def organization_set_id(cls) -> Mapped[UUID]:
        return create_mapped_column(
            DOMAIN, model.OrganizationSetMember, "organization_set_id"
        )

    @declared_attr
    def organization_id(cls) -> Mapped[UUID]:
        return create_mapped_column(
            DOMAIN, model.OrganizationSetMember, "organization_id"
        )

    @declared_attr
    def organization_set(cls) -> Mapped[model.OrganizationSet]:
        return relationship(
            "OrganizationSet",
            foreign_keys="OrganizationSetMember.organization_set_id",
        )

    @declared_attr
    def organization(cls) -> Mapped[model.Organization]:
        return relationship(
            "Organization", foreign_keys="OrganizationSetMember.organization_id"
        )


@declarative_mixin
class SiteMixin(RowMetadataMixin):
    @declared_attr
    def organization_id(cls) -> Mapped[UUID]:
        return create_mapped_column(DOMAIN, model.Site, "organization_id")

    name: Mapped[str] = create_mapped_column(DOMAIN, model.Site, "name")

    @declared_attr
    def organization(cls) -> Mapped[model.Organization]:
        return relationship("Organization", foreign_keys="Site.organization_id")


@declarative_mixin
class ContactMixin(RowMetadataMixin):
    @declared_attr
    def site_id(cls) -> Mapped[UUID | None]:
        return create_mapped_column(DOMAIN, model.Contact, "site_id")

    name: Mapped[str] = create_mapped_column(DOMAIN, model.Contact, "name")
    email: Mapped[str | None] = create_mapped_column(DOMAIN, model.Contact, "email")
    phone: Mapped[str | None] = create_mapped_column(DOMAIN, model.Contact, "phone")

    @declared_attr
    def site(cls) -> Mapped[SiteMixin | None]:
        return relationship("Site", foreign_keys="Contact.site_id")


@declarative_mixin
class IdentifierIssuerMixin(RowMetadataMixin):
    name: Mapped[str] = create_mapped_column(DOMAIN, model.IdentifierIssuer, "name")


@declarative_mixin
class DataCollectionMixin(RowMetadataMixin):
    name: Mapped[str] = create_mapped_column(DOMAIN, model.DataCollection, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DataCollection, "description"
    )


@declarative_mixin
class DataCollectionSetMixin(RowMetadataMixin):
    name: Mapped[str] = create_mapped_column(DOMAIN, model.DataCollectionSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.DataCollectionSet, "description"
    )


@declarative_mixin
class DataCollectionSetMemberMixin(RowMetadataMixin):
    @declared_attr
    def data_collection_set_id(cls) -> Mapped[UUID]:
        return create_mapped_column(
            DOMAIN,
            model.DataCollectionSetMember,
            "data_collection_set_id",
        )

    @declared_attr
    def data_collection_id(cls) -> Mapped[UUID]:
        return create_mapped_column(
            DOMAIN,
            model.DataCollectionSetMember,
            "data_collection_id",
        )

    @declared_attr
    def data_collection_set(cls) -> Mapped[DataCollectionSetMixin]:
        return relationship(
            "DataCollectionSet",
            foreign_keys="DataCollectionSetMember.data_collection_set_id",
        )

    @declared_attr
    def data_collection(cls) -> Mapped[DataCollectionMixin]:
        return relationship(
            "DataCollection", foreign_keys="DataCollectionSetMember.data_collection_id"
        )


@declarative_mixin
class UserInvitationMixin(RowMetadataMixin):
    key: Mapped[str] = create_mapped_column(DOMAIN, model.UserInvitation, "key")
    email: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.UserInvitation, "email"
    )
    name: Mapped[str | None] = create_mapped_column(
        DOMAIN, model.UserInvitation, "name"
    )
    token: Mapped[str] = create_mapped_column(DOMAIN, model.UserInvitation, "token")
    expires_at: Mapped[datetime.datetime] = create_mapped_column(
        DOMAIN, model.UserInvitation, "expires_at"
    )
    roles: Mapped[set[str]] = create_mapped_column(
        DOMAIN, model.UserInvitation, "roles"
    )

    @declared_attr
    def invited_by_user_id(cls) -> Mapped[UUID]:
        return create_mapped_column(DOMAIN, model.UserInvitation, "invited_by_user_id")

    @declared_attr
    def organization_id(cls) -> Mapped[UUID]:
        return create_mapped_column(DOMAIN, model.UserInvitation, "organization_id")

    @declared_attr
    def invited_by_user(cls) -> Mapped[UserMixin]:
        return relationship("User", foreign_keys="UserInvitation.invited_by_user_id")

    @declared_attr
    def organization(cls) -> Mapped[model.Organization]:
        return relationship(
            "Organization", foreign_keys="UserInvitation.organization_id"
        )


class Organization(Base, OrganizationMixin):
    __tablename__, __table_args__ = create_table_args(model.Organization)


class User(Base, UserMixin):
    __tablename__, __table_args__ = create_table_args(model.User)


class OrganizationSet(Base, OrganizationSetMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationSet)


class OrganizationSetMember(Base, OrganizationSetMemberMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationSetMember)


class Site(Base, SiteMixin):
    __tablename__, __table_args__ = create_table_args(model.Site)


class Contact(Base, ContactMixin):
    __tablename__, __table_args__ = create_table_args(model.Contact)


class IdentifierIssuer(Base, IdentifierIssuerMixin):
    __tablename__, __table_args__ = create_table_args(model.IdentifierIssuer)


class DataCollection(Base, DataCollectionMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollection)


class DataCollectionSet(Base, DataCollectionSetMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSet)


class DataCollectionSetMember(Base, DataCollectionSetMemberMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSetMember)


class UserInvitation(Base, UserInvitationMixin):
    __tablename__, __table_args__ = create_table_args(model.UserInvitation)



from typing import ClassVar
from uuid import UUID

from pydantic import Field, model_validator

from gen_epix.casedb.domain import enum
from gen_epix.casedb.domain.model.abac.rights import (
    CaseTypeAccessAbac,
    CaseTypeShareAbac,
)
from gen_epix.casedb.domain.model.case.non_persistable import CaseTypeDim
from gen_epix.casedb.domain.model.case.persistable import (
    CaseType,
    CaseTypeCol,
    Col,
    Dim,
    GeneticDistanceProtocol,
    TreeAlgorithm,
)
from gen_epix.casedb.domain.model.ontology import EtiologicalAgent, Etiology
from gen_epix.fastapp.domain import Entity


class CompleteCaseType(CaseType):
    NAME: ClassVar = "CompleteCaseType"
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_case_types",
        persistable=False,
    )
    etiologies: dict[UUID, Etiology] = Field(
        description="The etiologies used by the case type"
    )
    etiological_agents: dict[UUID, EtiologicalAgent] = Field(
        description="The etiological agents used by the case type"
    )
    dims: dict[UUID, Dim] = Field(description="The dimensions used by the case type")
    cols: dict[UUID, Col] = Field(description="The columns used by the case type")
    case_type_dims: list[CaseTypeDim] = Field(
        description="The ordered list of case type dimensions"
    )
    case_type_cols: dict[UUID, CaseTypeCol] = Field(
        description="The case type columns for the case type"
    )
    case_type_col_order: list[UUID] = Field(
        description="The order of the case type columns outside the context of a dimension"
    )
    genetic_distance_protocols: dict[UUID, GeneticDistanceProtocol] = Field(
        description="The genetic distance protocols used by the case type"
    )
    tree_algorithms: dict[enum.TreeAlgorithmType, TreeAlgorithm] = Field(
        description="The tree algorithms used by the case type"
    )
    case_type_access_abacs: dict[UUID, CaseTypeAccessAbac] = Field(
        description="The case type access ABAC object by data collection ID"
    )
    case_type_share_abacs: dict[UUID, CaseTypeShareAbac] = Field(
        description="The case type share ABAC object by data collection ID"
    )

    @model_validator(mode="after")
    def derive_case_type_col_order(self) -> "CompleteCaseType":
        ordered: list[UUID] = []
        seen: set[UUID] = set()
        for dim in self.case_type_dims:
            for cid in dim.case_type_col_order:
                if cid not in seen:
                    ordered.append(cid)
                    seen.add(cid)
        self.case_type_col_order = ordered
        return self



from typing import ClassVar

import gen_epix.casedb.domain.model as model
from gen_epix.commondb.domain.command import CrudCommand

# Non-CRUD


# CRUD


class OrganizationAccessCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationAccessCasePolicy


class UserAccessCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserAccessCasePolicy


class OrganizationShareCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.OrganizationShareCasePolicy


class UserShareCasePolicyCrudCommand(CrudCommand):
    MODEL_CLASS: ClassVar = model.UserShareCasePolicy



# pylint: disable=useless-import-alias
from gen_epix.commondb.domain.policy.abac import (
    BaseIsOrganizationAdminPolicy as BaseIsOrganizationAdminPolicy,
)
from gen_epix.commondb.domain.policy.abac import (
    BaseReadOrganizationResultsOnlyPolicy as BaseReadOrganizationResultsOnlyPolicy,
)
from gen_epix.commondb.domain.policy.abac import (
    BaseReadSelfResultsOnlyPolicy as BaseReadSelfResultsOnlyPolicy,
)
from gen_epix.commondb.domain.policy.abac import (
    BaseReadUserPolicy as BaseReadUserPolicy,
)
from gen_epix.commondb.domain.policy.abac import (
    BaseUpdateUserPolicy as BaseUpdateUserPolicy,
)
from gen_epix.commondb.domain.policy.permission import (
    NO_RBAC_PERMISSIONS as NO_RBAC_PERMISSIONS,
)
from gen_epix.commondb.domain.policy.permission import RoleGenerator as RoleGenerator
from gen_epix.commondb.domain.policy.rbac import (
    BaseIsPermissionSubsetNewRolePolicy as BaseIsPermissionSubsetNewRolePolicy,
)
from gen_epix.commondb.domain.policy.system import (
    BaseHasSystemOutagePolicy as BaseHasSystemOutagePolicy,
)



# pylint: disable=useless-import-alias
from gen_epix.fastapp.repositories.sa.mapper import SAMapper as SAMapper
from gen_epix.fastapp.repositories.sa.repository import SARepository as SARepository
from gen_epix.fastapp.repositories.sa.unit_of_work import SAUnitOfWork as SAUnitOfWork
from gen_epix.fastapp.repositories.sa.util import (
    ServerUtcCurrentTime as ServerUtcCurrentTime,
)
from gen_epix.fastapp.repositories.sa.util import (
    ServerUtcTimestamp as ServerUtcTimestamp,
)
from gen_epix.fastapp.repositories.sa.util import (
    create_sa_type_from_field_info as create_sa_type_from_field_info,
)



from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, declarative_mixin, mapped_column
from sqlalchemy_utils.types.uuid import UUIDType


@declarative_mixin
class DataLineageMixin:
    provenance_id: Mapped[UUID | None] = mapped_column(UUIDType(), nullable=True)
    source_traceback: Mapped[str | None] = mapped_column(sa.Unicode(255), nullable=True)
